"""Models for storing applied force field parameters."""

from __future__ import annotations

import ast
import json
from typing import TYPE_CHECKING, Annotated, Any

import numpy
from numpy.typing import ArrayLike
from openff.toolkit import Quantity
from openff.utilities.utilities import has_package, requires_package
from pydantic import (
    Field,
    PrivateAttr,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    WrapSerializer,
)
from pydantic.functional_validators import WrapValidator

from openff.interchange._annotations import _Quantity
from openff.interchange.exceptions import MissingParametersError
from openff.interchange.models import (
    LibraryChargeTopologyKey,
    PotentialKey,
    TopologyKey,
)
from openff.interchange.pydantic import _BaseModel

if TYPE_CHECKING:
    if has_package("jax"):
        from jax import Array
    else:
        type Array = Any  # type: ignore[no-redef]
else:
    type Array = ArrayLike


class Potential(_BaseModel):
    """Base class for storing applied parameters."""

    parameters: dict[str, _Quantity] = Field(dict())
    map_key: int | None = None

    def __hash__(self) -> int:
        return hash(tuple(self.parameters.values()))


class WrappedPotential(_BaseModel):
    """Model storing other Potential model(s) inside inner data."""

    _inner_data: dict[Potential, float] = PrivateAttr()

    def __init__(self, data: Potential | dict) -> None:
        # Needed to set some Pydantic magic, at least __pydantic_private__;
        # won't actually process the input here
        super().__init__()

        if isinstance(data, Potential):
            data = {data: 1.0}

        self._inner_data = data

    @property
    def parameters(self) -> dict[str, Quantity]:
        """Get the parameters as represented by the stored potentials and coefficients."""
        keys: set[str] = {param_key for pot in self._inner_data.keys() for param_key in pot.parameters.keys()}
        params = dict()
        for key in keys:
            params.update(
                {
                    key: sum(coeff * pot.parameters[key] for pot, coeff in self._inner_data.items()),
                },
            )
        return params

    def __repr__(self) -> str:
        return str(self._inner_data)


def validate_potential_or_wrapped_potential(
    v: Any,
    handler: ValidatorFunctionWrapHandler,
    info: ValidationInfo,
) -> Potential | WrappedPotential:
    """Validate the parameters field of a Potential object."""
    if info.mode == "json":
        if "parameters" in v:
            return Potential.model_validate(v)
        else:
            return WrappedPotential.model_validate(v)
    else:
        raise NotImplementedError(f"Validation mode {info.mode} not implemented.")


PotentialOrWrappedPotential = Annotated[
    Potential | WrappedPotential,
    WrapValidator(validate_potential_or_wrapped_potential),
]


def validate_key_map(v: Any, handler, info) -> dict:
    """Validate the key_map field of a Collection object."""
    from openff.interchange.models import (
        AngleKey,
        BondKey,
        ImproperTorsionKey,
        LibraryChargeTopologyKey,
        ProperTorsionKey,
        SingleAtomChargeTopologyKey,
    )

    tmp = dict()
    if info.mode in ("json", "python"):
        for key, val in v.items():
            val_dict = json.loads(val)

            key_class: type[TopologyKey]

            match val_dict["associated_handler"]:
                case "Bonds":
                    key_class = BondKey
                case "Angles":
                    key_class = AngleKey
                case "ProperTorsions":
                    key_class = ProperTorsionKey
                case "ImproperTorsions":
                    key_class = ImproperTorsionKey
                case "LibraryCharges":
                    key_class = LibraryChargeTopologyKey  # type: ignore[assignment]
                case (
                    "ToolkitAM1BCCHandler"
                    | "molecules_with_preset_charges"
                    | "NAGLChargesHandler"
                    | "ChargeIncrementModelHandler"
                ):
                    key_class = SingleAtomChargeTopologyKey  # type: ignore[assignment]
                case _:
                    key_class = TopologyKey

            try:
                tmp.update(
                    {
                        key_class.model_validate_json(
                            key,
                        ): PotentialKey.model_validate_json(val),
                    },
                )
            except Exception as error:
                raise ValueError(
                    f"Failed to deserialize a `PotentialKey` with {val_dict['associated_handler']=}",
                ) from error

            del key_class

        v = tmp

    else:
        raise ValueError(f"Validation mode {info.mode} not implemented.")

    return v


def serialize_key_map(
    value: dict[TopologyKey, PotentialKey],
    handler,
    info,
) -> dict[str, str]:
    """Serialize the parameters field of a Potential object."""
    if info.mode == "json":
        return {key.model_dump_json(): value.model_dump_json() for key, value in value.items()}

    else:
        raise NotImplementedError(f"Serialization mode {info.mode} not implemented.")


KeyMap = Annotated[
    dict[TopologyKey | LibraryChargeTopologyKey, PotentialKey],
    WrapValidator(validate_key_map),
    WrapSerializer(serialize_key_map),
]


def validate_potential_dict(
    v: Any,
    handler: ValidatorFunctionWrapHandler,
    info: ValidationInfo,
):
    """Validate the parameters field of a Potential object."""
    if info.mode == "json":
        return {PotentialKey.model_validate_json(key): Potential.model_validate_json(val) for key, val in v.items()}

    elif info.mode == "python":
        # Unclear why str sometimes sneak into here in Python mode; everything
        # should be object (PotentialKey/Potential) or dict at this point ...
        return {
            PotentialKey.model_validate_json(key) if isinstance(key, str) else key: (
                Potential.model_validate_json(val) if isinstance(val, str) else val
            )
            for key, val in v.items()
        }

    else:
        raise NotImplementedError(f"Validation mode {info.mode} not implemented.")


def serialize_potentials(
    value: dict[PotentialKey, Potential],
    handler,
    info,
) -> dict[str, str]:
    """Serialize the potentials field."""
    if info.mode == "json":
        return {key.model_dump_json(): _value.model_dump_json() for key, _value in value.items()}

    else:
        raise NotImplementedError(f"Serialization mode {info.mode} not implemented.")


Potentials = Annotated[
    dict[PotentialKey, PotentialOrWrappedPotential],
    WrapValidator(validate_potential_dict),
    WrapSerializer(serialize_potentials),
]


class Collection(_BaseModel):
    """Base class for storing parametrized force field data."""

    type: str = Field(..., description="The type of potentials this handler stores.")
    is_plugin: bool = Field(
        False,
        description="Whether this collection is defined as a plugin.",
    )
    expression: str = Field(
        ...,
        description="The analytical expression governing the potentials in this handler.",
    )
    key_map: KeyMap = Field(
        dict(),
        description="A mapping between TopologyKey objects and PotentialKey objects.",
    )
    potentials: Potentials = Field(
        dict(),
        description="A mapping between PotentialKey objects and Potential objects.",
    )

    @property
    def independent_variables(self) -> set[str]:
        """
        Return a set of variables found in the expression but not in any potentials.
        """
        vars_in_potentials = next(iter(self.potentials.values())).parameters.keys()
        vars_in_expression = {node.id for node in ast.walk(ast.parse(self.expression)) if isinstance(node, ast.Name)}
        return vars_in_expression - vars_in_potentials

    def _get_parameters(self, atom_indices: tuple[int]) -> dict:
        for topology_key in self.key_map:
            if topology_key.atom_indices == atom_indices:
                potential_key = self.key_map[topology_key]
                potential = self.potentials[potential_key]
                parameters = potential.parameters
                return parameters
        raise MissingParametersError(
            f"Could not find parameter in parameter in handler {self.type} associated with atoms {atom_indices}",
        )

    def get_force_field_parameters(
        self,
        use_jax: bool = False,
    ) -> ArrayLike | Array:
        """Return a flattened representation of the force field parameters."""
        # TODO: Handle WrappedPotential
        if any(isinstance(potential, WrappedPotential) for potential in self.potentials.values()):
            raise NotImplementedError

        if use_jax:
            from jax import numpy as jax_numpy

            return jax_numpy.array(
                [[v.m for v in p.parameters.values()] for p in self.potentials.values()],
            )
        else:
            return numpy.array(
                [[v.m for v in p.parameters.values()] for p in self.potentials.values()],
            )

    def set_force_field_parameters(self, new_p: ArrayLike) -> None:
        """Set the force field parameters from a flattened representation."""

        mapping = self.get_mapping()
        if new_p.shape[0] != len(mapping):  # type: ignore
            raise RuntimeError

        for potential_key, potential_index in self.get_mapping().items():
            potential = self.potentials[potential_key]
            if len(new_p[potential_index, :]) != len(potential.parameters):  # type: ignore
                raise RuntimeError

            for parameter_index, parameter_key in enumerate(potential.parameters):
                parameter_units = potential.parameters[parameter_key].units
                modified_parameter = new_p[potential_index, parameter_index]  # type: ignore

                self.potentials[potential_key].parameters[parameter_key] = modified_parameter * parameter_units

    def get_system_parameters(
        self,
        p=None,
        use_jax: bool = False,
    ) -> ArrayLike | Array:
        """
        Return a flattened representation of system parameters.

        These values are effectively force field parameters as applied to a chemical topology.
        """
        # TODO: Handle WrappedPotential
        if any(isinstance(potential, WrappedPotential) for potential in self.potentials.values()):
            raise NotImplementedError

        if p is None:
            p = self.get_force_field_parameters(use_jax=use_jax)
        mapping = self.get_mapping()

        q: list = list()
        for potential_key in self.key_map.values():
            index = mapping[potential_key]
            q.append(p[index])

        if use_jax:
            from jax import numpy as jax_numpy

            return jax_numpy.array(q)
        else:
            return numpy.array(q)

    def get_mapping(self) -> dict[PotentialKey, int]:
        """Get a mapping between potentials and array indices."""
        mapping: dict = dict()
        index = 0
        for potential_key in self.key_map.values():
            if potential_key not in mapping:
                mapping[potential_key] = index
                index += 1

        return mapping

    def parametrize(
        self,
        p=None,
        use_jax: bool = True,
    ) -> ArrayLike | Array:
        """Return an array of system parameters, given an array of force field parameters."""
        if p is None:
            p = self.get_force_field_parameters(use_jax=use_jax)

        return self.get_system_parameters(p=p, use_jax=use_jax)

    def parametrize_partial(self):
        """Return a function that will call `self.parametrize()` with arguments specified by `self.mapping`."""
        from functools import partial

        return partial(
            self.parametrize,
            mapping=self.get_mapping(),
        )

    @requires_package("jax")
    def get_param_matrix(self) -> ArrayLike | Array:
        """Get a matrix representing the mapping between force field and system parameters."""
        from functools import partial

        import jax

        p = self.get_force_field_parameters(use_jax=True)

        parametrize_partial = partial(
            self.parametrize,
        )

        jac_parametrize = jax.jacfwd(parametrize_partial)
        jac_res = jac_parametrize(p)

        return jac_res.reshape(-1, p.flatten().shape[0])  # type: ignore[union-attr]

    def __getitem__(self, key) -> Potential | WrappedPotential:
        if isinstance(key, tuple) and key not in self.key_map and tuple(reversed(key)) in self.key_map:
            return self.potentials[self.key_map[tuple(reversed(key))]]  # type: ignore[index]

        return self.potentials[self.key_map[key]]

    def __hash__(self) -> int:
        return hash(id(self))

    def __eq__(self, other) -> bool:
        if type(self) is not type(other):
            return NotImplemented

        return self is other


def validate_collections(
    v: Any,
    handler: ValidatorFunctionWrapHandler,
    info: ValidationInfo,
) -> dict:
    """Validate the collections dict from a JSON blob."""
    from openff.interchange.smirnoff import (
        SMIRNOFFAngleCollection,
        SMIRNOFFBondCollection,
        SMIRNOFFCollection,
        SMIRNOFFConstraintCollection,
        SMIRNOFFElectrostaticsCollection,
        SMIRNOFFImproperTorsionCollection,
        SMIRNOFFProperTorsionCollection,
        SMIRNOFFvdWCollection,
        SMIRNOFFVirtualSiteCollection,
    )

    _class_mapping: dict[str, type[SMIRNOFFCollection]] = {
        "Bonds": SMIRNOFFBondCollection,
        "Angles": SMIRNOFFAngleCollection,
        "Constraints": SMIRNOFFConstraintCollection,
        "ProperTorsions": SMIRNOFFProperTorsionCollection,
        "ImproperTorsions": SMIRNOFFImproperTorsionCollection,
        "vdW": SMIRNOFFvdWCollection,
        "Electrostatics": SMIRNOFFElectrostaticsCollection,
        "VirtualSites": SMIRNOFFVirtualSiteCollection,
    }

    if info.mode in ("json", "python"):
        return {
            collection_name: _class_mapping[collection_name].model_validate(
                collection_data,
            )
            for collection_name, collection_data in v.items()
        }
    else:
        raise ValueError(f"Validation mode {info.mode} not implemented.")


_AnnotatedCollections = Annotated[
    dict[str, Collection],
    WrapValidator(validate_collections),
]



from collections.abc import Generator
from typing import Any

import numpy
import pandas

class Element:
    atomic_number: int

class Atom:
    name: str
    _bond_partners: list[Atom]
    index: int
    element: Element
    serial: int | None

class Bond:
    atom1: Atom
    atom2: Atom

class Chain: ...

class Residue:
    name: str
    chain: Chain
    resSeq: int | None
    segment_id: str | None
    @property
    def atoms(self) -> Generator[Atom]: ...

class Topology:
    n_atoms: int
    @classmethod
    def from_openmm(cls, value: Any) -> Topology: ...
    def to_openmm(self) -> Any: ...
    def add_chain(self) -> Any: ...
    def add_residue(
        self,
        name: str,
        chain: Any,
        resSeq: int | None = None,
        segment_id: str | None = None,
    ) -> Any: ...
    def add_atom(
        self,
        name: str,
        element: Any,
        residue: Any,
        serial: int | None = None,
    ): ...
    def add_bond(
        self,
        atom1: Any,
        atom2: Any,
        type: Any | None = None,
        order: int | None = None,
    ): ...
    def atom(self, index: int) -> Atom: ...
    @property
    def atoms(self) -> Generator[Atom]: ...
    @property
    def bonds(self) -> Generator[Bond]: ...
    @property
    def residues(self) -> Generator[Residue]: ...
    def subset(self, atom_indices: numpy.ndarray | list[int]) -> Topology: ...
    def join(self, other: Topology, keep_resSeq: bool = True) -> Topology: ...
    def to_dataframe(self) -> tuple[pandas.DataFrame, numpy.ndarray]: ...
    def residue(self, index: int) -> Residue: ...

class Trajectory:
    topology: Topology
    n_atoms: int
    xyz: numpy.ndarray
    unitcell_lengths: numpy.ndarray
    unitcell_angles: numpy.ndarray
    unitcell_vectors: numpy.ndarray

    def save_pdb(self, file: str): ...

def load_pdb(file: str) -> Trajectory: ...
def load(file: str) -> Trajectory: ...



import abc
from typing import Self, TypeVar

from openff.toolkit import Topology
from openff.toolkit.topology import Atom, ValenceDict
from openff.toolkit.typing.engines.smirnoff.parameters import (
    AngleHandler,
    BondHandler,
    ImproperTorsionHandler,
    ParameterHandler,
    ProperTorsionHandler,
)

from openff.interchange.components.potentials import Collection
from openff.interchange.exceptions import (
    InvalidParameterHandlerError,
    SMIRNOFFParameterAttributeNotImplementedError,
    UnassignedAngleError,
    UnassignedBondError,
    UnassignedTorsionError,
)
from openff.interchange.models import (
    LibraryChargeTopologyKey,
    PotentialKey,
    TopologyKey,
)

TP = TypeVar("TP", bound="ParameterHandler")


# Coped from the toolkit, see
# https://github.com/openforcefield/openff-toolkit/blob/0133414d3ab51e1af0996bcebe0cc1bdddc6431b/
# openff/toolkit/typing/engines/smirnoff/parameters.py#L2318
def _check_all_valence_terms_assigned(
    handler_class: type,
    assigned_atom_indices: set[tuple[int, ...]],
    topology: Topology,
    valence_terms: list[tuple[Atom, ...]],
):
    """Check that all valence terms have been assigned."""
    if len(assigned_atom_indices) == len(valence_terms):
        return

    # Convert the valence term to a valence dictionary to make sure
    # the order of atom indices doesn't matter for comparison.
    valence_terms_dict = (
        ValenceDict()
    )  # might be some cases in which a subclass is used, but we're only using this for key ordering ...
    for atoms in valence_terms:
        atom_indices = (topology.atom_index(a) for a in atoms)
        valence_terms_dict[atom_indices] = None

    # Check that both valence dictionaries have the same keys (i.e. terms).
    assigned_terms_set = set(assigned_atom_indices)
    valence_terms_set = set(valence_terms_dict.keys())
    unassigned_terms = valence_terms_set.difference(assigned_terms_set)
    not_found_terms = assigned_terms_set.difference(valence_terms_set)

    # Raise an error if there are unassigned terms.
    err_msg = ""

    unassigned_atom_tuples = []
    if len(unassigned_terms) > 0:
        unassigned_str = ""
        for unassigned_tuple in unassigned_terms:
            unassigned_str += "\n- Topology indices " + str(unassigned_tuple)
            unassigned_str += ": names and elements "

            unassigned_atoms = []

            # Pull and add additional helpful info on missing terms
            for atom_idx in unassigned_tuple:
                atom = topology.atom(atom_idx)
                unassigned_atoms.append(atom)
                unassigned_str += f"({atom.name} {atom.symbol}), "
            unassigned_atom_tuples.append(tuple(unassigned_atoms))

        err_msg += (
            f"{handler_class.__name__} was not able to find parameters for the "
            f"following valence terms:\n{unassigned_str}"
        )

    if len(not_found_terms) > 0:
        if err_msg != "":
            err_msg += "\n"
        not_found_str = "\n- ".join([str(x) for x in not_found_terms])
        err_msg += f"{handler_class.__name__} assigned terms that were not found in the topology:\n- {not_found_str}"
    if err_msg:
        err_msg += "\n"

        class_exception_map = {
            BondHandler: UnassignedBondError,
            AngleHandler: UnassignedAngleError,
            ProperTorsionHandler: UnassignedTorsionError,
            ImproperTorsionHandler: UnassignedTorsionError,
        }

        try:
            exception = class_exception_map[handler_class](err_msg)
        except KeyError:
            raise RuntimeError(
                f"Could not find an exception class for handler with class {handler_class}",
            )

        exception.unassigned_topology_atom_tuples = unassigned_atom_tuples  # type: ignore[attr-defined]
        exception.handler_class = handler_class  # type: ignore[attr-defined]

        raise exception


class SMIRNOFFCollection(Collection, abc.ABC):
    """Base class for handlers storing potentials produced by SMIRNOFF force fields."""

    type: str

    is_plugin: bool = False

    def modify_openmm_forces(self, *args, **kwargs):
        """Optionally modify, create, or delete forces. Currently only available to plugins."""
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def supported_parameters(cls):
        """Return a list of parameter attributes supported by this handler."""
        raise NotImplementedError()

    @classmethod
    def potential_parameters(cls):
        """Return a subset of `supported_parameters` that are meant to be included in potentials."""
        raise NotImplementedError()

    @classmethod
    def check_supported_parameters(cls, parameter_handler: ParameterHandler):
        """Verify that a parameter handler is in an allowed list of handlers."""
        for parameter in parameter_handler.parameters:
            for parameter_attribute in parameter._get_defined_parameter_attributes():
                if parameter_attribute == "parent_id":
                    continue
                if parameter_attribute not in cls.supported_parameters():
                    raise SMIRNOFFParameterAttributeNotImplementedError(
                        parameter_attribute,
                    )

    @classmethod
    def check_openmm_requirements(cls, combine_nonbonded_forces: bool) -> None:
        """Run through a list of assertions about what is compatible when exporting this to OpenMM."""

    def store_matches(
        self,
        parameter_handler: ParameterHandler,
        topology: "Topology",
    ) -> None:
        """Populate self.key_map with key-val pairs of [TopologyKey, PotentialKey]."""
        if self.key_map:
            # TODO: Should the key_map always be reset, or should we be able to partially
            # update it? Also Note the duplicated code in the child classes
            self.key_map: dict[
                TopologyKey | LibraryChargeTopologyKey,
                PotentialKey,
            ] = dict()

        matches = parameter_handler.find_matches(topology)

        for key, val in matches.items():
            parameter: ParameterHandler.ParameterType = val.parameter_type

            cosmetic_attributes = {
                cosmetic_attribute: getattr(
                    parameter,
                    f"_{cosmetic_attribute}",
                )
                for cosmetic_attribute in parameter._cosmetic_attribs
            }

            topology_key = TopologyKey(atom_indices=key)

            potential_key = PotentialKey(
                id=parameter.smirks,
                associated_handler=parameter_handler.TAGNAME,
                cosmetic_attributes=cosmetic_attributes,
            )

            self.key_map[topology_key] = potential_key

    def store_potentials(self, parameter_handler: TP):
        """
        Populate self.potentials with key-val pairs of [PotentialKey, Potential].
        """
        raise NotImplementedError()

    @classmethod
    def create(
        cls,
        parameter_handler: TP,
        topology: "Topology",
    ) -> Self:
        """
        Create a SMIRNOFFCOllection from toolkit data.

        """
        if type(parameter_handler) not in cls.allowed_parameter_handlers():
            raise InvalidParameterHandlerError(type(parameter_handler))

        collection = cls()  # type: ignore[call-arg]
        if hasattr(collection, "fractional_bondorder_method"):
            if getattr(parameter_handler, "fractional_bondorder_method", None):
                collection.fractional_bond_order_method = (  # type: ignore[attr-defined]
                    parameter_handler.fractional_bondorder_method
                )
                collection.fractional_bond_order_interpolation = (  # type: ignore[attr-defined]
                    parameter_handler.fractional_bondorder_interpolation
                )
        collection.store_matches(parameter_handler=parameter_handler, topology=topology)
        collection.store_potentials(parameter_handler=parameter_handler)

        return collection

    def __repr__(self) -> str:
        return (
            f"Handler '{self.type}' with expression '{self.expression}', {len(self.key_map)} mapping keys, "
            f"and {len(self.potentials)} potentials"
        )



import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.utilities.testing import skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import get_test_file_path
from openff.interchange.constants import kcal_mol, kcal_mol_a2
from openff.interchange.exceptions import (
    PresetChargesError,
    UnassignedAngleError,
    UnassignedBondError,
    UnassignedTorsionError,
)
from openff.interchange.models import BondKey
from openff.interchange.smirnoff import (
    SMIRNOFFAngleCollection,
    SMIRNOFFBondCollection,
    SMIRNOFFvdWCollection,
)
from openff.interchange.smirnoff._nonbonded import (
    _upconvert_vdw_handler,
    library_charge_from_molecule,
)
from openff.interchange.warnings import PresetChargesAndVirtualSitesWarning


def _get_interpolated_bond_k(bond_handler) -> float:
    for key in bond_handler.key_map:
        if key.bond_order is not None:
            topology_key = key
            break
    potential_key = bond_handler.key_map[topology_key]
    return bond_handler.potentials[potential_key].parameters["k"].m


class TestCreate:
    def test_modified_nonbonded_cutoffs(self, sage, ethanol):
        topology = Topology.from_molecules(ethanol)
        modified_sage = ForceField(sage.to_string())

        modified_sage["vdW"].cutoff = Quantity(0.777, "angstrom")
        modified_sage["Electrostatics"].cutoff = Quantity(0.777, "angstrom")

        out = Interchange.from_smirnoff(force_field=modified_sage, topology=topology)

        assert out["vdW"].cutoff == Quantity(0.777, "angstrom")
        assert out["Electrostatics"].cutoff == Quantity(0.777, "angstrom")

    def test_sage_tip3p_charges(self, water, sage):
        """Ensure tip3p charges packaged with sage are applied over AM1-BCC charges.
        https://github.com/openforcefield/openff-toolkit/issues/1199"""
        out = Interchange.from_smirnoff(force_field=sage, topology=[water])
        found_charges = [v.m for v in out["Electrostatics"]._get_charges().values()]

        assert numpy.allclose(found_charges, [-0.834, 0.417, 0.417])

    def test_infer_positions(self, sage, ethanol):
        assert Interchange.from_smirnoff(sage, [ethanol]).positions is None

        ethanol.generate_conformers(n_conformers=1)

        assert Interchange.from_smirnoff(sage, [ethanol]).positions.shape == (
            ethanol.n_atoms,
            3,
        )

    def test_cosmetic_attributes(self):
        from openff.toolkit._tests.test_forcefield import xml_ff_w_cosmetic_elements

        force_field = ForceField(
            "openff-2.1.0.offxml",
            xml_ff_w_cosmetic_elements.replace(
                'Bonds version="0.3"',
                'Bonds version="0.4"',
            ),
            allow_cosmetic_attributes=True,
        )

        bonds = SMIRNOFFBondCollection()

        bonds.store_matches(
            parameter_handler=force_field["Bonds"],
            topology=Molecule.from_smiles("CC").to_topology(),
        )

        for key in bonds.potentials:
            if key.id == "[#6X4:1]-[#6X4:2]":
                assert key.cosmetic_attributes == {
                    "parameters": "k, length",
                    "parameterize_eval": "blah=blah2",
                }

    def test_all_cosmetic(self, sage, basic_top):
        for handler in sage.registered_parameter_handlers:
            for parameter in sage[handler].parameters:
                parameter._cosmetic_attribs = ["fOO"]
                parameter._fOO = "bAR"
                parameter.fOO = "bAR"

        out = sage.create_interchange(basic_top)

        for collection in out.collections:
            if collection == "Electrostatics":
                continue

            for potential_key in out[collection].potentials:
                assert potential_key.cosmetic_attributes["fOO"] == "bAR"


class TestUnassignedParameters:
    def test_catch_unassigned_bonds(self, sage, ethanol_top):
        for param in sage["Bonds"].parameters:
            param.smirks = "[#99:1]-[#99:2]"

        sage.deregister_parameter_handler(sage["Constraints"])

        with pytest.raises(
            UnassignedBondError,
            match="BondHandler was not able to find par",
        ):
            Interchange.from_smirnoff(force_field=sage, topology=ethanol_top)

    def test_catch_unassigned_angles(self, sage, ethanol_top):
        for param in sage["Angles"].parameters:
            param.smirks = "[#99:1]-[#99:2]-[#99:3]"

        with pytest.raises(
            UnassignedAngleError,
            match="AngleHandler was not able to find par",
        ):
            Interchange.from_smirnoff(force_field=sage, topology=ethanol_top)

    def test_catch_unassigned_torsions(self, sage, ethanol_top):
        for param in sage["ProperTorsions"].parameters:
            param.smirks = "[#99:1]-[#99:2]-[#99:3]-[#99:4]"

        with pytest.raises(
            UnassignedTorsionError,
            match="- Topology indices [(]5, 0, 1, 6[)]: "
            r"names and elements [(](H\d+)? H[)], [(](C\d+)? C[)], [(](C\d+)? C[)], [(](H\d+)? H[)],",
        ):
            Interchange.from_smirnoff(force_field=sage, topology=ethanol_top)


# TODO: Remove xfail after openff-toolkit 0.10.0
@pytest.mark.xfail
def test_library_charges_from_molecule():
    from openff.toolkit.typing.engines.smirnoff.parameters import LibraryChargeHandler

    mol = Molecule.from_mapped_smiles("[Cl:1][C:2]#[C:3][F:4]")

    with pytest.raises(ValueError, match="missing partial"):
        library_charge_from_molecule(mol)

    mol.partial_charges = Quantity(numpy.linspace(-0.3, 0.3, 4), "elementary_charge")

    library_charges = library_charge_from_molecule(mol)

    assert isinstance(library_charges, LibraryChargeHandler.LibraryChargeType)
    assert library_charges.smirks == mol.to_smiles(mapped=True)
    assert library_charges.charge == [*mol.partial_charges]


class TestPresetCharges:
    @pytest.mark.slow
    def test_charge_from_molecules_basic(self, sage):
        molecule = Molecule.from_smiles("CCO")
        molecule.assign_partial_charges(partial_charge_method="am1bcc")
        molecule.partial_charges *= -1

        default = Interchange.from_smirnoff(sage, molecule.to_topology())
        uses = Interchange.from_smirnoff(
            sage,
            molecule.to_topology(),
            charge_from_molecules=[molecule],
        )

        found_charges_no_uses = [v.m for v in default["Electrostatics"]._get_charges().values()]
        found_charges_uses = [v.m for v in uses["Electrostatics"]._get_charges().values()]

        assert not numpy.allclose(found_charges_no_uses, found_charges_uses)

        assert numpy.allclose(found_charges_uses, molecule.partial_charges.m)

    def test_charges_on_molecules_in_topology(self, sage, water):
        ethanol = Molecule.from_smiles("CCO")

        ethanol_charges = numpy.linspace(-1, 1, 9) * 0.4
        water_charges = numpy.linspace(-1, 1, 3)

        ethanol.partial_charges = Quantity(ethanol_charges, "elementary_charge")
        water.partial_charges = Quantity(water_charges, "elementary_charge")

        out = Interchange.from_smirnoff(
            sage,
            [ethanol],
            charge_from_molecules=[ethanol, water],
        )

        for molecule in out.topology.molecules:
            if "C" in molecule.to_smiles():
                assert numpy.allclose(molecule.partial_charges.m, ethanol_charges)
            else:
                assert numpy.allclose(molecule.partial_charges.m, water_charges)

    def test_charges_from_molecule_reordered(
        self,
        sage,
        hydrogen_cyanide,
        hydrogen_cyanide_reversed,
    ):
        """Test the behavior of charge_from_molecules when the atom ordering differs with the topology"""

        # H - C # N
        molecule = hydrogen_cyanide

        #  N  # C  - H
        # -0.3, 0.0, 0.3
        molecule_with_charges = hydrogen_cyanide_reversed
        molecule_with_charges.partial_charges = Quantity(
            [-0.3, 0.0, 0.3],
            "elementary_charge",
        )

        out = Interchange.from_smirnoff(
            sage,
            molecule.to_topology(),
            charge_from_molecules=[molecule_with_charges],
        )

        expected_charges = [0.3, 0.0, -0.3]
        found_charges = [v.m for v in out["Electrostatics"]._get_charges().values()]

        assert numpy.allclose(expected_charges, found_charges)

    def test_error_when_any_missing_partial_charges(self, sage):
        ethanol = Molecule.from_smiles("CCO")

        ethanol.assign_partial_charges(partial_charge_method="am1bcc")

        topology = Topology.from_molecules(
            [
                Molecule.from_smiles("C"),
                ethanol,
            ],
        )

        with pytest.raises(
            PresetChargesError,
            match=r"All.*must have partial charges",
        ):
            sage.create_interchange(
                topology,
                charge_from_molecules=[Molecule.from_smiles("C")],
            )

    def test_error_multiple_isomorphic_molecules(self, sage, ethanol, reversed_ethanol):
        # these ethanol fixtures *do* have charges, if they didn't have charges it would be
        # ambiguous which error should be raised
        topology = Topology.from_molecules([ethanol, reversed_ethanol, ethanol])

        with pytest.raises(
            PresetChargesError,
            match=r"All molecules.*isomorphic.*unique",
        ):
            sage.create_interchange(
                topology,
                charge_from_molecules=[ethanol, reversed_ethanol],
            )

    def test_virtual_site_charge_increments_applied_after_preset_charges_water(
        self,
        tip4p,
        water,
    ):
        """
        Test that charge increments to/from virtual sites are still applied after preset charges are set.

        This is funky user input, since preset charges override library charges, which are important for water.
        """
        water.partial_charges = Quantity(
            [-2.0, 1.0, 1.0],
            "elementary_charge",
        )

        with pytest.warns(PresetChargesAndVirtualSitesWarning):
            # This dict is probably ordered O H H VS
            charges = tip4p.create_interchange(
                water.to_topology(),
                charge_from_molecules=[water],
            )["Electrostatics"].charges

        # tip4p_fb.offxml is meant to result in charges of -1.0517362213526 (VS) 0 (O) and 0.5258681106763 (H)
        # the force field has 0.0 for all library charges (skipping AM1-BCC), so preset charges screw this up.
        # the resulting charges, if using preset charges of [-2, 1, 1], should be the result of adding together
        # [-2, 1, 1, 0] (preset charges) and
        # [0, 1.5258681106763001, 1.5258681106763001, -1.0517362213526] (charge increments)

        numpy.testing.assert_allclose(
            [charge.m for charge in charges.values()],
            [-2.0, 1.5258681106763001, 1.5258681106763001, -1.0517362213526],
        )

    def test_virtual_site_charge_increments_applied_after_preset_charges_ligand(
        self,
        sage_with_sigma_hole,
    ):
        ligand = Molecule.from_mapped_smiles("[C:1]([Cl:2])([Cl:3])([Cl:4])[Cl:5]")

        ligand.partial_charges = Quantity(
            [1.2, -0.3, -0.3, -0.3, -0.3],
            "elementary_charge",
        )

        with pytest.warns(PresetChargesAndVirtualSitesWarning):
            # This dict is probably ordered C Cl Cl Cl Cl VS VS VS VS
            charges = sage_with_sigma_hole.create_interchange(
                ligand.to_topology(),
                charge_from_molecules=[ligand],
            )["Electrostatics"].charges

        # this fake sigma hole parameter shifts 0.1 e from the carbon and 0.2 e from the chlorine, so the
        # resulting charges should be like
        # [1.2, -0.3, -0.3, -0.3, -0.3] +
        # [0.4, 0.2, 0.2, 0.2, 0.2, -0.3, -0.3, -0.3, -0.3]

        numpy.testing.assert_allclose(
            [charge.m for charge in charges.values()],
            [1.6, -0.1, -0.1, -0.1, -0.1, -0.3, -0.3, -0.3, -0.3],
        )


@skip_if_missing("openmm")
class TestPartialBondOrdersFromMolecules:
    @pytest.mark.parametrize(
        (
            "reversed",
            "central_atoms",
        ),
        [
            (False, (1, 2)),
            (True, (7, 6)),
        ],
    )
    def test_interpolated_partial_bond_orders_from_molecules(
        self,
        reversed,
        central_atoms,
        ethanol,
        reversed_ethanol,
    ):
        """Test the fractional bond orders are used to interpolate k and length values as we expect,
        including that the fractional bond order is defined by the value on the input molecule via
        `partial_bond_order_from_molecules`, not whatever is produced by a default call to
        `Molecule.assign_fractional_bond_orders`.

        This test is adapted from test_fractional_bondorder_from_molecule in the toolkit.

        Parameter   | param values at bond orders 1, 2  | used bond order   | expected value
        bond k        101, 123 kcal/mol/A**2              1.55                113.1 kcal/mol/A**2
        bond length   1.4, 1.3 A                          1.55                1.345 A
        torsion k     1, 1.8 kcal/mol                     1.55                1.44 kcal/mol
        """
        from openff.toolkit._tests.test_forcefield import xml_ff_bo

        molecule = reversed_ethanol if reversed else ethanol

        molecule.get_bond_between(*central_atoms).fractional_bond_order = 1.55

        sorted_indices = tuple(sorted(central_atoms))

        out = Interchange.from_smirnoff(
            force_field=ForceField(
                "openff-2.0.0.offxml",
                xml_ff_bo,
            ),
            topology=molecule.to_topology(),
            partial_bond_orders_from_molecules=[molecule],
        )

        bond_key = BondKey(atom_indices=sorted_indices, bond_order=1.55)
        bond_potential = out["Bonds"].key_map[bond_key]
        found_bond_k = out["Bonds"].potentials[bond_potential].parameters["k"]
        found_bond_length = out["Bonds"].potentials[bond_potential].parameters["length"]

        assert found_bond_k.m_as(kcal_mol_a2) == pytest.approx(113.1)
        assert found_bond_length.m_as("angstrom") == pytest.approx(1.345)

        # TODO: There should be a better way of plucking this torsion's TopologyKey
        for topology_key in out["ProperTorsions"].key_map.keys():
            if (tuple(sorted(topology_key.atom_indices))[1:3] == sorted_indices) and topology_key.bond_order == 1.55:
                torsion_key = topology_key
                break

        torsion_potential = out["ProperTorsions"].key_map[torsion_key]
        found_torsion_k = out["ProperTorsions"].potentials[torsion_potential].parameters["k"]

        assert found_torsion_k.m_as(kcal_mol) == pytest.approx(1.44)

    @pytest.mark.slow
    def test_partial_bond_order_from_molecules_empty(self, ethanol):
        from openff.toolkit._tests.test_forcefield import xml_ff_bo

        forcefield = ForceField(
            "openff-2.0.0.offxml",
            xml_ff_bo,
        )

        default = Interchange.from_smirnoff(forcefield, ethanol.to_topology())
        empty = Interchange.from_smirnoff(
            forcefield,
            ethanol.to_topology(),
            partial_bond_orders_from_molecules=list(),
        )

        assert _get_interpolated_bond_k(default["Bonds"]) == pytest.approx(
            _get_interpolated_bond_k(empty["Bonds"]),
        )

    def test_partial_bond_order_from_molecules_no_matches(self, ethanol):
        from openff.toolkit._tests.test_forcefield import xml_ff_bo

        forcefield = ForceField(
            "openff-2.0.0.offxml",
            xml_ff_bo,
        )

        decoy = Molecule.from_smiles("C#N")
        decoy.assign_fractional_bond_orders(bond_order_model="am1-wiberg")

        default = Interchange.from_smirnoff(forcefield, ethanol.to_topology())
        uses = Interchange.from_smirnoff(
            forcefield,
            ethanol.to_topology(),
            partial_bond_orders_from_molecules=[decoy],
        )

        assert _get_interpolated_bond_k(default["Bonds"]) == pytest.approx(
            _get_interpolated_bond_k(uses["Bonds"]),
        )


class TestCreateWithPlugins:
    def test_setup_plugins(self):
        from nonbonded_plugins.nonbonded import (
            BuckinghamHandler,
            SMIRNOFFBuckinghamCollection,
        )

        from openff.interchange.smirnoff._create import _PLUGIN_CLASS_MAPPING

        assert _PLUGIN_CLASS_MAPPING[BuckinghamHandler] == SMIRNOFFBuckinghamCollection

    def test_create_buckingham(self, water):
        force_field = ForceField(
            get_test_file_path("buckingham.offxml"),
            load_plugins=True,
        )

        out = Interchange.from_smirnoff(
            force_field,
            water.to_topology(),
        )

        assert "Buckingham" in out.collections
        assert len(out["Buckingham"].potentials) == 2


@skip_if_missing("jax")
class TestMatrixRepresentations:
    @pytest.mark.parametrize(
        ("handler_name", "n_ff_terms", "n_sys_terms"),
        [("vdW", 10, 72), ("Bonds", 8, 64), ("Angles", 6, 104)],
    )
    def test_to_force_field_to_system_parameters(
        self,
        sage,
        ethanol_top,
        handler_name,
        n_ff_terms,
        n_sys_terms,
    ):
        import jax

        if handler_name == "Bonds":
            handler = SMIRNOFFBondCollection.create(
                parameter_handler=sage["Bonds"],
                topology=ethanol_top,
            )
        elif handler_name == "Angles":
            handler = SMIRNOFFAngleCollection.create(
                parameter_handler=sage[handler_name],
                topology=ethanol_top,
            )
        elif handler_name == "vdW":
            # Ensure this is 0.3 until everything assumes 0.4
            _upconvert_vdw_handler(sage[handler_name])

            handler = SMIRNOFFvdWCollection.create(
                parameter_handler=sage[handler_name],
                topology=ethanol_top,
            )
        else:
            raise NotImplementedError()

        p = handler.get_force_field_parameters(use_jax=True)

        assert isinstance(p, jax.Array)
        assert numpy.prod(p.shape) == n_ff_terms

        q = handler.get_system_parameters(use_jax=True)

        assert isinstance(q, jax.Array)
        assert numpy.prod(q.shape) == n_sys_terms

        assert jax.numpy.allclose(q, handler.parametrize(p))

        param_matrix = handler.get_param_matrix()

        ref_file = get_test_file_path(f"ethanol_param_{handler_name.lower()}.npy")
        ref = jax.numpy.load(ref_file)

        assert jax.numpy.allclose(ref, param_matrix)

        # TODO: Update with other handlers that can safely be assumed to follow 1:1 slot:smirks mapping
        if handler_name in ["vdW", "Bonds", "Angles"]:
            assert numpy.allclose(
                numpy.sum(param_matrix, axis=1),
                numpy.ones(param_matrix.shape[0]),
            )

    def test_set_force_field_parameters(self, sage, ethanol):
        import jax

        bond_handler = SMIRNOFFBondCollection.create(
            parameter_handler=sage["Bonds"],
            topology=ethanol.to_topology(),
        )

        original = bond_handler.get_force_field_parameters(use_jax=True)
        modified = original * jax.numpy.array([1.1, 0.5])

        bond_handler.set_force_field_parameters(modified)

        assert (bond_handler.get_force_field_parameters(use_jax=True) == modified).all()



import warnings
from typing import Literal, Self

from openff.toolkit import Molecule, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff.parameters import (
    AngleHandler,
    BondHandler,
    ConstraintHandler,
    ImproperTorsionHandler,
    ParameterHandler,
    ProperTorsionHandler,
)
from openff.toolkit.utils.exceptions import SMIRNOFFSpecError
from packaging.version import Version

from openff.interchange.common._valence import (
    AngleCollection,
    BondCollection,
    ImproperTorsionCollection,
    ProperTorsionCollection,
)
from openff.interchange.components.potentials import Potential, WrappedPotential
from openff.interchange.components.toolkit import _cache_angle_parameter_lookup, _cache_torsion_parameter_lookup
from openff.interchange.exceptions import (
    DuplicateMoleculeError,
    InvalidParameterHandlerError,
    MissingParametersError,
)
from openff.interchange.models import (
    BondKey,
    ImproperTorsionKey,
    PotentialKey,
    ProperTorsionKey,
)
from openff.interchange.smirnoff._base import (
    SMIRNOFFCollection,
)
from openff.interchange.warnings import ForceFieldModificationWarning


def _upconvert_bondhandler(bond_handler: BondHandler):
    """Given a BondHandler with version 0.3, up-convert to 0.4."""
    if bond_handler.version >= Version("0.4"):
        return

    elif bond_handler.version == Version("0.3"):
        warnings.warn(
            "Automatically up-converting BondHandler from version 0.3 to 0.4. Consider manually upgrading "
            "this BondHandler (or <Bonds> section in an OFFXML file) to 0.4 or newer. For more details, "
            "see https://openforcefield.github.io/standards/standards/smirnoff/#bonds.",
            ForceFieldModificationWarning,
        )

        bond_handler.version = Version("0.4")
        bond_handler.potential = "(k/2)*(r-length)^2"


def _check_molecule_uniqueness(molecule_list: list[Molecule] | None):
    """Check if the reference molecule is isomorphic with any molecules in a provided list."""
    # TODO: This could all be replaced by MoleculeSet
    if molecule_list is None:
        return

    for index, molecule in enumerate(molecule_list):
        for other_index, other_molecule in enumerate(molecule_list):
            if other_index <= index:
                continue
            if other_molecule.is_isomorphic_with(molecule):
                # The toolkit used to enforce that `partial_bond_orders_from_molecules` must not have isomorphic
                # duplicates in its list, raising `ValueError` if any fail.

                raise DuplicateMoleculeError(
                    "Duplicate molecules found in `partial_bond_orders_from_molecules` list. "
                    "Please ensure that each molecule in this list is isomorphically unique.",
                )


def _molecule_is_in_list(
    molecule: Molecule,
    molecule_list: list[Molecule] | None,
) -> bool:
    if molecule_list is None:
        return False

    for list_molecule in molecule_list:
        if molecule.is_isomorphic_with(list_molecule):
            return True

    return False


def _get_interpolation_coeffs(fractional_bond_order, data):
    x1, x2 = data.keys()
    coeff1 = (x2 - fractional_bond_order) / (x2 - x1)
    coeff2 = (fractional_bond_order - x1) / (x2 - x1)

    return coeff1, coeff2


class SMIRNOFFBondCollection(SMIRNOFFCollection, BondCollection):
    """Collection storing bond potentials as produced by a SMIRNOFF force field."""

    type: Literal["Bonds"] = "Bonds"
    expression: Literal["k/2*(r-length)**2"] = "k/2*(r-length)**2"
    fractional_bond_order_method: Literal["AM1-Wiberg", "None", "none"] = "AM1-Wiberg"
    fractional_bond_order_interpolation: Literal["linear"] = "linear"

    @classmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        return [BondHandler]

    @classmethod
    def supported_parameters(cls):
        """Return a list of supported parameter attribute names."""
        return ["smirks", "id", "k", "length", "k_bondorder", "length_bondorder"]

    @classmethod
    def potential_parameters(cls):
        """Return a list of names of parameters included in each potential in this colletion."""
        return ["k", "length"]

    @classmethod
    def valence_terms(cls, topology):
        """Return all bonds in this topology."""
        return [tuple(b.atoms) for b in topology.bonds]

    def store_matches(
        self,
        parameter_handler: ParameterHandler,
        topology: Topology,
    ) -> None:
        """
        Populate self.key_map with key-val pairs of keys and unique potential identifiers.
        """
        if self.key_map:
            # TODO: Should the key_map always be reset, or should we be able to partially
            # update it? Also Note the duplicated code in the child classes
            self.key_map: dict[BondKey, PotentialKey] = dict()
        matches = parameter_handler.find_matches(topology)
        for key, val in matches.items():
            parameter: BondHandler.BondType = val.parameter_type

            cosmetic_attributes = {
                cosmetic_attribute: getattr(
                    parameter,
                    f"_{cosmetic_attribute}",
                )
                for cosmetic_attribute in parameter._cosmetic_attribs
            }

            if parameter.k_bondorder or parameter.length_bondorder:
                bond = topology.get_bond_between(*key)
                fractional_bond_order = bond.fractional_bond_order
                if not fractional_bond_order:
                    assert self._get_uses_interpolation(parameter_handler)
                    raise RuntimeError(
                        "Bond orders should already be assigned at this point",
                    )
            else:
                fractional_bond_order = None
            topology_key = BondKey(
                atom_indices=key,
                bond_order=fractional_bond_order,
            )

            potential_key = PotentialKey(
                id=parameter.smirks,
                associated_handler=parameter_handler.TAGNAME,
                bond_order=fractional_bond_order,
                cosmetic_attributes=cosmetic_attributes,
            )

            self.key_map[topology_key] = potential_key

    def store_potentials(self, parameter_handler: BondHandler) -> None:
        """
        Populate self.potentials with key-val pairs of [PotentialKey, Potential].

        """
        if self.potentials:
            self.potentials = dict()
        for topology_key, potential_key in self.key_map.items():
            smirks = potential_key.id
            force_field_parameters = parameter_handler.parameters[smirks]

            if topology_key.bond_order:  # type: ignore[union-attr]
                bond_order = topology_key.bond_order  # type: ignore[union-attr]
                if force_field_parameters.k_bondorder:
                    data = force_field_parameters.k_bondorder
                else:
                    data = force_field_parameters.length_bondorder
                coeffs = _get_interpolation_coeffs(
                    fractional_bond_order=bond_order,
                    data=data,
                )
                pots = []
                map_keys = [*data.keys()]
                for map_key in map_keys:
                    pots.append(
                        Potential(
                            parameters={
                                parameter: getattr(
                                    force_field_parameters,
                                    parameter + "_bondorder",
                                )[map_key]
                                for parameter in self.potential_parameters()
                            },
                            map_key=map_key,
                        ),
                    )

                potential: Potential | WrappedPotential = WrappedPotential(
                    {pot: coeff for pot, coeff in zip(pots, coeffs)},
                )

            else:
                potential = Potential(
                    parameters={
                        parameter: getattr(force_field_parameters, parameter)
                        for parameter in self.potential_parameters()
                    },
                )

            self.potentials[potential_key] = potential

    # This could probably be a static or class method if it's used heavily, or maybe
    # even a standalone function overloaded to the different handlers it takes in
    def _get_uses_interpolation(self, parameter_handler: BondHandler) -> bool:
        for parameter in parameter_handler.parameters:
            if parameter.k_bondorder is not None:
                return True
            if parameter.length_bondorder is not None:
                return True

        return False

    @classmethod
    def create(
        cls,
        parameter_handler: BondHandler,
        topology: Topology,
        partial_bond_orders_from_molecules: list[Molecule] | None = None,
    ) -> Self:
        """
        Create a SMIRNOFFBondCollection from toolkit data.

        """
        # TODO: This method overrides SMIRNOFFCollection.from_toolkit in order to gobble up
        # a ConstraintHandler. This seems like a good solution for the interdependence, but is also
        # not a great practice. A better solution would involve not overriding the method with a
        # different function signature.
        if type(parameter_handler) not in cls.allowed_parameter_handlers():
            raise InvalidParameterHandlerError

        collection = cls(
            type="Bonds",
            expression="k/2*(r-length)**2",
            fractional_bond_order_method=parameter_handler.fractional_bondorder_method,
            fractional_bond_order_interpolation=parameter_handler.fractional_bondorder_interpolation,
        )

        if collection._get_uses_interpolation(parameter_handler):
            _check_molecule_uniqueness(partial_bond_orders_from_molecules)

            for molecule in topology.molecules:
                # TODO: This loop could be sped up by iterating over `unique_molecules` and later copy the
                if _molecule_is_in_list(molecule, partial_bond_orders_from_molecules):
                    continue

                # TODO: expose conformer generation and fractional bond order assigment knobs to user via API
                molecule.generate_conformers(n_conformers=1)
                molecule.assign_fractional_bond_orders(
                    bond_order_model=collection.fractional_bond_order_method.lower(),
                )

        collection.store_matches(parameter_handler=parameter_handler, topology=topology)
        collection.store_potentials(parameter_handler=parameter_handler)

        return collection


class SMIRNOFFConstraintCollection(SMIRNOFFCollection):
    """Handler storing constraint potentials as produced by a SMIRNOFF force field."""

    type: Literal["Constraints"] = "Constraints"
    expression: Literal[""] = ""

    @classmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        return [BondHandler, ConstraintHandler]

    @classmethod
    def supported_parameters(cls):
        """Return a list of supported parameter attribute names."""
        return ["smirks", "id", "length", "distance"]

    @classmethod
    def potential_parameters(cls):
        """Return a list of names of parameters included in each potential in this colletion."""
        return ["length", "distance"]

    @classmethod
    def create(
        cls,
        parameter_handler: list,
        topology: Topology,
        bonds: SMIRNOFFBondCollection | None = None,
    ) -> Self:
        """
        Create a SMIRNOFFCollection from toolkit data.

        """
        if isinstance(parameter_handler, list):
            parameter_handlers = parameter_handler
        else:
            parameter_handlers = [parameter_handler]

        for parameter_handler in parameter_handlers:
            if type(parameter_handler) not in cls.allowed_parameter_handlers():
                raise InvalidParameterHandlerError(type(parameter_handler))

        collection = cls()
        collection.store_constraints(
            parameter_handlers=parameter_handlers,
            topology=topology,
            bonds=bonds,
        )

        return collection

    def store_constraints(
        self,
        parameter_handlers: list,
        topology: Topology,
        bonds: SMIRNOFFBondCollection | None = None,
    ) -> None:
        """Store constraints."""
        if self.key_map:
            self.key_map = dict()

        try:
            constraint_handler = next(p for p in parameter_handlers if type(p) is ConstraintHandler)
        except (StopIteration, IndexError):
            return

        constraint_matches = constraint_handler.find_matches(topology)

        if any([type(p) is BondHandler for p in parameter_handlers]):
            bond_handler = next(p for p in parameter_handlers if type(p) is BondHandler)
            # This should be passed in as a parameter
            assert bonds is not None
        else:
            bond_handler = None
            bonds = None

        for key, match in constraint_matches.items():
            topology_key = BondKey(atom_indices=key)

            parameter = match.parameter_type

            smirks = parameter.smirks
            distance = parameter.distance
            cosmetic_attributes = {
                cosmetic_attribute: getattr(
                    parameter,
                    f"_{cosmetic_attribute}",
                )
                for cosmetic_attribute in parameter._cosmetic_attribs
            }

            if distance is not None:
                # This constraint parameter is fully specified
                potential_key = PotentialKey(
                    id=smirks,
                    associated_handler="Constraints",
                    cosmetic_attributes=cosmetic_attributes,
                )

                self.key_map[topology_key] = potential_key

                distance = parameter.distance

            else:
                # This constraint parameter depends on the BondHandler ...
                if bond_handler is None:
                    raise MissingParametersError(
                        f"Constraint with SMIRKS pattern {smirks} found with no distance "
                        "specified, and no corresponding bond parameters were found. The distance "
                        "of this constraint is not specified.",
                    )

                try:
                    # ... so use the same PotentialKey instance as the BondHandler to look up the distance
                    potential_key = bonds.key_map[topology_key]  # type: ignore[union-attr]
                except KeyError:
                    raise SMIRNOFFSpecError(
                        "Error while processing a constraint with no distance specified and no harmonic "
                        f"bond parameters associated with atoms {topology_key.atom_indices=}. This is "
                        "unsupported in the SMIRNOFF specification. Consider adding a distance to the "
                        "constraint parameter or instead constraining other atoms which are bonded and "
                        "have harmonic bond parameters.",
                    )

                self.key_map[topology_key] = potential_key

                distance = bonds.potentials[potential_key].parameters["length"]  # type: ignore[union-attr]

            potential = Potential(
                parameters={
                    "distance": distance,
                },
            )

            self.potentials[potential_key] = potential


class SMIRNOFFAngleCollection(SMIRNOFFCollection, AngleCollection):
    """Handler storing angle potentials as produced by a SMIRNOFF force field."""

    type: Literal["Angles"] = "Angles"
    expression: Literal["k/2*(theta-angle)**2"] = "k/2*(theta-angle)**2"

    @classmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        return [AngleHandler]

    @classmethod
    def supported_parameters(cls):
        """Return a list of supported parameter attributes."""
        return ["smirks", "id", "k", "angle"]

    @classmethod
    def potential_parameters(cls):
        """Return a list of names of parameters included in each potential in this colletion."""
        return ["k", "angle"]

    @classmethod
    def valence_terms(cls, topology):
        """Return all angles in this topology."""
        return list(topology.angles)

    def store_potentials(self, parameter_handler: AngleHandler) -> None:
        """
        Populate self.potentials with key-val pairs of [PotentialKey, Potential].

        """
        for potential_key in self.key_map.values():
            self.potentials.update(
                {
                    potential_key: Potential(
                        parameters=_cache_angle_parameter_lookup(
                            potential_key,
                            parameter_handler,
                        ),
                    ),
                },
            )


class SMIRNOFFProperTorsionCollection(SMIRNOFFCollection, ProperTorsionCollection):
    """Handler storing proper torsions potentials as produced by a SMIRNOFF force field."""

    type: Literal["ProperTorsions"] = "ProperTorsions"
    expression: Literal["k*(1+cos(periodicity*theta-phase))"] = "k*(1+cos(periodicity*theta-phase))"
    fractional_bond_order_method: Literal["AM1-Wiberg"] = "AM1-Wiberg"
    fractional_bond_order_interpolation: Literal["linear"] = "linear"

    @classmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        return [ProperTorsionHandler]

    @classmethod
    def supported_parameters(cls):
        """Return a list of supported parameter attribute names."""
        return ["smirks", "id", "k", "periodicity", "phase", "idivf", "k_bondorder"]

    @classmethod
    def potential_parameters(cls):
        """Return a list of names of parameters included in each potential in this colletion."""
        return ["k", "periodicity", "phase", "idivf"]

    @classmethod
    def valence_terms(cls, topology):
        """Return all (proper) torsions in this topology."""
        return list(topology.propers)

    def store_matches(
        self,
        parameter_handler: ProperTorsionHandler,
        topology: Topology,
    ) -> None:
        """
        Populate self.key_map with key-val pairs of keys and unique potential identifiers.

        """
        if self.key_map:
            self.key_map: dict[ProperTorsionKey, PotentialKey] = dict()
        matches = parameter_handler.find_matches(topology)
        for key, val in matches.items():
            parameter: ProperTorsionHandler.ProperTorsionType = val.parameter_type

            n_terms = len(parameter.phase)

            cosmetic_attributes = {
                cosmetic_attribute: getattr(
                    parameter,
                    f"_{cosmetic_attribute}",
                )
                for cosmetic_attribute in parameter._cosmetic_attribs
            }

            for n in range(n_terms):
                smirks = parameter.smirks

                if parameter.k_bondorder:
                    # The relevant bond order is that of the _central_ bond in the torsion
                    bond = topology.get_bond_between(key[1], key[2])
                    fractional_bond_order = bond.fractional_bond_order
                    if not fractional_bond_order:
                        raise RuntimeError(
                            "Bond orders should already be assigned at this point",
                        )
                else:
                    fractional_bond_order = None

                topology_key = ProperTorsionKey(
                    atom_indices=key,
                    mult=n,
                    bond_order=fractional_bond_order,
                )

                potential_key = PotentialKey(
                    id=smirks,
                    mult=n,
                    associated_handler="ProperTorsions",
                    bond_order=fractional_bond_order,
                    cosmetic_attributes=cosmetic_attributes,
                )

                self.key_map[topology_key] = potential_key

    def store_potentials(self, parameter_handler: ProperTorsionHandler) -> None:
        """
        Populate self.potentials with key-val pairs of [PotentialKey, Potential].

        """
        for topology_key, potential_key in self.key_map.items():
            if topology_key.bond_order:  # type: ignore[union-attr]
                smirks = potential_key.id
                n = potential_key.mult
                parameter = parameter_handler.parameters[smirks]

                bond_order = topology_key.bond_order  # type: ignore[union-attr]
                data = parameter.k_bondorder[n]
                coeffs = _get_interpolation_coeffs(
                    fractional_bond_order=bond_order,
                    data=data,
                )
                pots = []
                map_keys = [*data.keys()]
                for map_key in map_keys:
                    parameters = {
                        "k": parameter.k_bondorder[n][map_key],
                        "periodicity": Quantity(parameter.periodicity[n], "dimensionless"),
                        "phase": parameter.phase[n],
                        "idivf": Quantity(parameter.idivf[n], "dimensionless"),
                    }
                    pots.append(
                        Potential(
                            parameters=parameters,
                            map_key=map_key,
                        ),
                    )
                potential = WrappedPotential(
                    {pot: coeff for pot, coeff in zip(pots, coeffs)},
                )
            else:
                parameters = _cache_torsion_parameter_lookup(potential_key, parameter_handler)
                potential = Potential(parameters=parameters)  # type: ignore[assignment]
            self.potentials[potential_key] = potential

    @classmethod
    def create(
        cls,
        parameter_handler: ProperTorsionHandler,
        topology: Topology,
        partial_bond_orders_from_molecules=None,
    ) -> Self:
        """
        Create a SMIRNOFFProperTorsionCollection from toolkit data.

        """
        collection = cls(
            type="ProperTorsions",
            expression="k*(1+cos(periodicity*theta-phase))",
            fractional_bond_order_method=parameter_handler.fractional_bondorder_method,
            fractional_bond_order_interpolation=parameter_handler.fractional_bondorder_interpolation,
        )

        if any(getattr(p, "k_bondorder", None) is not None for p in parameter_handler.parameters):
            _check_molecule_uniqueness(partial_bond_orders_from_molecules)

            for molecule in topology.molecules:
                if _molecule_is_in_list(molecule, partial_bond_orders_from_molecules):
                    continue

                # TODO: expose conformer generation and fractional bond order assigment knobs via API?
                molecule.generate_conformers(n_conformers=1)
                molecule.assign_fractional_bond_orders(
                    bond_order_model=collection.fractional_bond_order_method.lower(),
                )

        collection.store_matches(parameter_handler=parameter_handler, topology=topology)
        collection.store_potentials(parameter_handler=parameter_handler)

        return collection


class SMIRNOFFImproperTorsionCollection(SMIRNOFFCollection, ImproperTorsionCollection):
    """Handler storing improper torsions potentials as produced by a SMIRNOFF force field."""

    type: Literal["ImproperTorsions"] = "ImproperTorsions"
    expression: Literal["k*(1+cos(periodicity*theta-phase))"] = "k*(1+cos(periodicity*theta-phase))"
    # TODO: Consider whether or not default_idivf should be stored here

    @classmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        return [ImproperTorsionHandler]

    @classmethod
    def supported_parameters(cls):
        """Return a list of supported parameter attribute names."""
        return ["smirks", "id", "k", "periodicity", "phase", "idivf"]

    @classmethod
    def potential_parameters(cls):
        """Return a list of names of parameters included in each potential in this colletion."""
        return ["k", "periodicity", "phase", "idivf"]

    def store_matches(
        self,
        parameter_handler: ImproperTorsionHandler,
        topology: Topology,
    ) -> None:
        """
        Populate self.key_map with key-val pairs of keys and unique potential identifiers.

        """
        if self.key_map:
            self.key_map = dict()
        matches = parameter_handler.find_matches(topology)
        for key, val in matches.items():
            parameter_handler._assert_correct_connectivity(
                val,
                [
                    (0, 1),
                    (1, 2),
                    (1, 3),
                ],
            )

            parameter: ImproperTorsionHandler.ImproperTorsionType = val.parameter_type

            n_terms = len(parameter.phase)

            cosmetic_attributes = {
                cosmetic_attribute: getattr(
                    parameter,
                    f"_{cosmetic_attribute}",
                )
                for cosmetic_attribute in parameter._cosmetic_attribs
            }

            for n in range(n_terms):
                smirks = parameter.smirks

                non_central_indices = [key[0], key[2], key[3]]

                for permuted_key in [
                    (
                        non_central_indices[i],
                        non_central_indices[j],
                        non_central_indices[k],
                    )
                    for (i, j, k) in [(0, 1, 2), (1, 2, 0), (2, 0, 1)]
                ]:
                    topology_key = ImproperTorsionKey(
                        atom_indices=(key[1], *permuted_key),
                        mult=n,
                    )
                    potential_key = PotentialKey(
                        id=smirks,
                        mult=n,
                        associated_handler="ImproperTorsions",
                        cosmetic_attributes=cosmetic_attributes,
                    )

                    self.key_map[topology_key] = potential_key

    def store_potentials(self, parameter_handler: ImproperTorsionHandler) -> None:
        """
        Populate self.potentials with key-val pairs of [PotentialKey, Potential].

        """
        _default_idivf = parameter_handler.default_idivf

        for potential_key in self.key_map.values():
            smirks = potential_key.id
            n = potential_key.mult
            parameter = parameter_handler.parameters[smirks]
            if parameter.idivf is None:
                idivf = None
            else:
                # Assumed to be list here
                idivf = parameter.idivf[n]
                if idivf is not None:
                    idivf = Quantity(idivf, "dimensionless")

            if idivf is None:
                if _default_idivf == "auto":
                    idivf = Quantity(3.0, "dimensionless")
                else:
                    # Assumed to be a numerical value
                    idivf = Quantity(_default_idivf, "dimensionless")

            # parameter keys happen to be the same as keys in proper torsions
            self.potentials[potential_key] = Potential(
                parameters=_cache_torsion_parameter_lookup(
                    potential_key,
                    parameter_handler,
                    idivf=idivf,
                ),
            )



from openff.utilities.utilities import has_package

from openff.interchange import Interchange
from openff.interchange.exceptions import UnsupportedExportError
from openff.interchange.smirnoff._gbsa import SMIRNOFFGBSACollection

if has_package("openmm"):
    import openmm


def _process_gbsa(
    interchange: "Interchange",
    system: "openmm.System",
):
    import openmm.app

    try:
        collection: SMIRNOFFGBSACollection = interchange.collections["GBSA"]  # type: ignore[assignment]
    except KeyError:
        return

    existing_forces = [
        force for force in system.getForces() if isinstance(force, openmm.NonbondedForce | openmm.CustomNonbondedForce)
    ]

    if len(existing_forces) != 1:
        raise UnsupportedExportError(
            "GBSA implementation assumes exactly one NonbondedForce or CustomNonbondedForce is present."
            f"Found these ({len(existing_forces)=}) forces: {existing_forces=}",
        )

    non_bonded_force = existing_forces[0]

    if non_bonded_force.getNonbondedMethod() == openmm.NonbondedForce.NoCutoff:
        amber_cutoff = None
    else:
        amber_cutoff = non_bonded_force.getCutoffDistance()

    if collection.gb_model == "OBC2":
        force = openmm.GBSAOBCForce()
    elif collection.gb_model in ["OBC1", "HCT"]:
        if collection.gb_model == "HCT":
            force_type = openmm.app.internal.customgbforces.GBSAHCTForce
        elif collection.gb_model == "OBC1":
            force_type = openmm.app.internal.customgbforces.GBSAOBC1Force

        force = force_type(
            solventDielectric=collection.solvent_dielectric,
            soluteDielectric=collection.solute_dielectric,
            SA=collection.sa_model,
            cutoff=amber_cutoff,
            kappa=0,
        )

    system.addForce(force)

    if amber_cutoff is not None:
        force.setCutoffDistance(amber_cutoff)

    if non_bonded_force.usesPeriodicBoundaryConditions():
        force.setNonbondedMethod(openmm.NonbondedForce.CutoffPeriodic)
    else:
        force.setNonbondedMethod(openmm.NonbondedForce.NoCutoff)

    if collection.gb_model == "OBC2":
        force.setSolventDielectric(collection.solvent_dielectric)
        force.setSoluteDielectric(collection.solute_dielectric)
        force.setSurfaceAreaEnergy(
            (0 if collection.sa_model is None else collection.surface_area_penalty.to_openmm()),
        )

    for topology_key, potential_key in collection.key_map.items():
        charge, *_ = non_bonded_force.getParticleParameters(
            topology_key.atom_indices[0],
        )
        _parameters = collection.potentials[potential_key].parameters

        parameters = [
            charge,
            _parameters["radius"].m_as("nanometer"),
            _parameters["scale"],
        ]

        if collection.gb_model == "OBC2":
            force.addParticle(*parameters)
        else:
            force.addParticle(parameters)

    if collection.gb_model != "OBC2":
        force.finalize()



"""
Collections for custom interactions.
"""


def load_smirnoff_plugins() -> list:
    """Load external potential handlers as plugins."""
    from importlib.metadata import entry_points

    from pydantic import PydanticUserError

    plugin_handlers = list()

    for entry_point in entry_points().select(
        group="openff.interchange.plugins.collections",
    ):
        try:
            plugin_handlers.append(entry_point.load())
        except PydanticUserError:
            continue

    return plugin_handlers



"""
Helper functions for exporting the topology to OpenMM.
"""

from typing import TYPE_CHECKING

from openff.toolkit import Topology
from openff.utilities import has_package

from openff.interchange import Interchange
from openff.interchange.models import BaseVirtualSiteKey

if has_package("openmm") or TYPE_CHECKING:
    import openmm.app


def to_openmm_topology(
    interchange: "Interchange",
    collate: bool = False,
    ensure_unique_atom_names: str | bool = "residues",
) -> "openmm.app.Topology":
    """
    Create an OpenMM Topology containing some virtual site information (if appropriate).

    Parameters
    ----------
    interchange
        The Interchange object to convert to an OpenMM Topology.
    collate
        If False, the default, virtual sites will be added to residues at the end of the topology.
        If True, virtual sites will be collated with their associated molecule and added to the residue of the last
        atom in the molecule they belong to.

    """
    # Heavily cribbed from the toolkit
    # https://github.com/openforcefield/openff-toolkit/blob/0.11.0rc2/openff/toolkit/topology/topology.py

    from openff.toolkit.topology import Bond
    from openff.toolkit.topology._mm_molecule import _SimpleBond

    from openff.interchange.interop._virtual_sites import _get_molecule_virtual_site_map

    # Copy topology to avoid modifying input (eg, when generating atom names)
    topology = Topology(interchange.topology)

    molecule_virtual_site_map = _get_molecule_virtual_site_map(interchange)

    has_virtual_sites = len(molecule_virtual_site_map) > 0

    virtual_site_element = openmm.app.element.Element.getByMass(0)

    openmm_topology = openmm.app.Topology()

    # Create unique atom names (as requested)
    if ensure_unique_atom_names:
        for molecule in topology._molecules:
            if isinstance(ensure_unique_atom_names, str) and hasattr(
                molecule,
                ensure_unique_atom_names,
            ):
                for hier_elem in getattr(molecule, ensure_unique_atom_names):
                    if not hier_elem.has_unique_atom_names:
                        hier_elem.generate_unique_atom_names()
            elif not molecule.has_unique_atom_names:
                molecule.generate_unique_atom_names()

    # Go through atoms in OpenFF to preserve the order.
    omm_atoms = []

    # For each atom in each molecule, determine which chain/residue it should be a part of
    for molecule in topology.molecules:
        molecule_index = topology.molecule_index(molecule)

        # No chain or residue can span more than one OFF molecule, so reset these to None for the first
        # atom in each molecule.
        last_chain = None
        last_residue = None
        for atom in molecule.atoms:
            # If the these are undefined, assume a default of
            # residue name/number "UNK" / "0", chain ID "X"
            atom_residue_name = atom.metadata.get("residue_name", "UNK")
            atom_residue_number = atom.metadata.get("residue_number", "0")
            atom_chain_id = atom.metadata.get("chain_id", "X")

            # Determine whether this atom should be part of the last atom's chain, or if it
            # should start a new chain
            if last_chain is None:
                chain = openmm_topology.addChain(id=atom_chain_id)
            elif last_chain.id == atom_chain_id:
                chain = last_chain
            else:
                chain = openmm_topology.addChain(id=atom_chain_id)
            # Determine whether this atom should be a part of the last atom's residue, or if it
            # should start a new residue
            if last_residue is None:
                residue = openmm_topology.addResidue(name=atom_residue_name, chain=chain)
                residue.id = atom_residue_number
            elif all(
                (
                    (last_residue.name == atom_residue_name),
                    (int(last_residue.id) == int(atom_residue_number)),
                    (chain.id == last_chain.id),  # type: ignore[union-attr]
                ),
            ):
                residue = last_residue
            else:
                residue = openmm_topology.addResidue(name=atom_residue_name, chain=chain)
                residue.id = atom_residue_number

            # Add atom.
            element = openmm.app.Element.getByAtomicNumber(atom.atomic_number)
            omm_atom = openmm_topology.addAtom(atom.name, element, residue)

            # Make sure that OpenFF and OpenMM Topology atoms have the same indices.
            # assert topology.atom_index(atom) == int(omm_atom.id) - 1
            omm_atoms.append(omm_atom)

            last_chain = chain
            last_residue = residue

        if has_virtual_sites and collate:
            virtual_sites_in_this_molecule: list[BaseVirtualSiteKey] = molecule_virtual_site_map[molecule_index]
            for this_virtual_site in virtual_sites_in_this_molecule:
                virtual_site_name = this_virtual_site.name

                # For now, assume that the residue of the last atom in the molecule is the same
                # residue as the entire molecule - this in unsafe for (bio)polymers/macromolecules
                virtual_site_residue = residue

                openmm_topology.addAtom(
                    virtual_site_name,
                    virtual_site_element,
                    virtual_site_residue,
                )

        # Add all bonds.
        bond_types = {1: openmm.app.Single, 2: openmm.app.Double, 3: openmm.app.Triple}
        for bond in molecule.bonds:
            atom1, atom2 = bond.atoms
            atom1_idx, atom2_idx = (
                topology.atom_index(atom1),
                topology.atom_index(
                    atom2,
                ),
            )
            if isinstance(bond, Bond):
                if bond.is_aromatic:
                    bond_type = openmm.app.Aromatic
                else:
                    bond_type = bond_types[bond.bond_order]
                bond_order = bond.bond_order
            elif isinstance(bond, _SimpleBond):
                bond_type = None
                bond_order = None
            else:
                raise RuntimeError(
                    "Unexpected bond type found while iterating over Topology.bonds."
                    f"Found {type(bond)}, allowed is Bond.",
                )

            openmm_topology.addBond(
                omm_atoms[atom1_idx],
                omm_atoms[atom2_idx],
                type=bond_type,
                order=bond_order,
            )

    if has_virtual_sites and not collate:
        chain = None
        residue = None
        for virtual_site_key in interchange["VirtualSites"].key_map:
            assert isinstance(virtual_site_key, BaseVirtualSiteKey)

            parent_atom_index = virtual_site_key.orientation_atom_indices[0]
            parent_atom = omm_atoms[parent_atom_index]

            if chain is None or chain.id != parent_atom.residue.chain.id:
                chain = openmm_topology.addChain(id=parent_atom.residue.chain.id)

            if residue is None or not (
                residue.name == parent_atom.residue.name
                and residue.id == parent_atom.residue.id
                and residue.insertionCode == parent_atom.residue.insertionCode
                and residue.chain.id == parent_atom.residue.chain.id
            ):
                residue = openmm_topology.addResidue(
                    name=parent_atom.residue.name,
                    chain=chain,
                    id=parent_atom.residue.id,
                    insertionCode=parent_atom.residue.insertionCode,
                )

            virtual_site_name = virtual_site_key.name

            openmm_topology.addAtom(
                name=virtual_site_name,
                element=virtual_site_element,
                residue=residue,
            )

    if interchange.box is not None:
        from openff.units.openmm import to_openmm

        openmm_topology.setPeriodicBoxVectors(to_openmm(interchange.box))

    return openmm_topology



"""Helpers for testing GROMACS interoperability."""

from openff.utilities import temporary_cd

from openff.interchange import Interchange
from openff.interchange.components.mdconfig import get_smirnoff_defaults
from openff.interchange.drivers import get_gromacs_energies


def gmx_roundtrip(state: Interchange, apply_smirnoff_defaults: bool = False):
    with temporary_cd():
        state.to_gromacs(prefix="state", decimal=8)
        new_state = Interchange.from_gromacs(topology_file="state.top", gro_file="state.gro")

        get_smirnoff_defaults(periodic=True).apply(new_state)

        get_gromacs_energies(state).compare(get_gromacs_energies(new_state))


def gmx_monolithic_vs_itp(state: Interchange):
    with temporary_cd():
        # TODO: More helpful handling of failures, i.e.
        #         * Detect differences in positions
        #         * Detect differences in box vectors
        #         * Detect differences in non-bonded settings
        get_gromacs_energies(state, _monolithic=True).compare(get_gromacs_energies(state, _monolithic=False))



from collections import defaultdict
from typing import TYPE_CHECKING

from openff.utilities.utilities import has_package, requires_package

if TYPE_CHECKING or has_package("openmm"):
    import openmm
    import openmm.unit

    kj_nm2_mol = openmm.unit.kilojoule_per_mole / openmm.unit.nanometer**2
    kj_rad2_mol = openmm.unit.kilojoule_per_mole / openmm.unit.radian**2


@requires_package("openmm")
def get_14_scaling_factors(omm_sys: "openmm.System") -> tuple[list, list]:
    """Find the 1-4 scaling factors as they are applied to an OpenMM System."""
    nonbond_force = next(f for f in omm_sys.getForces() if type(f) is openmm.NonbondedForce)

    vdw_14 = list()
    coul_14 = list()

    for exception_idx in range(nonbond_force.getNumExceptions()):
        i, j, q, _, eps = nonbond_force.getExceptionParameters(exception_idx)

        # Trust that q == 0 covers the cases of 1-2, 1-3, and truly being 0
        if q._value != 0:
            q_i = nonbond_force.getParticleParameters(i)[0]
            q_j = nonbond_force.getParticleParameters(j)[0]
            coul_14.append(q / (q_i * q_j))

        if eps._value != 0:
            eps_i = nonbond_force.getParticleParameters(i)[2]
            eps_j = nonbond_force.getParticleParameters(j)[2]
            vdw_14.append(eps / (eps_i * eps_j) ** 0.5)

    return coul_14, vdw_14


def _create_torsion_dict(torsion_force) -> dict[tuple[int], list[tuple]]:
    torsions = defaultdict(list)

    for i in range(torsion_force.getNumTorsions()):
        p1, p2, p3, p4, periodicity, phase, k = torsion_force.getTorsionParameters(i)
        key = (p1, p2, p3, p4)
        torsions[key]
        torsions[key].append((periodicity, phase, k))

    return torsions


def _create_bond_dict(bond_force):
    bonds = dict()

    for i in range(bond_force.getNumBonds()):
        p1, p2, length, k = bond_force.getBondParameters(i)
        key = (p1, p2)
        bonds[key] = (length, k)

    return bonds


def _create_angle_dict(angle_force):
    angles = dict()

    for i in range(angle_force.getNumAngles()):
        p1, p2, p3, theta, k = angle_force.getAngleParameters(i)
        key = (p1, p2, p3)
        angles[key] = (theta, k)

    return angles


@requires_package("openmm")
def _compare_individual_torsions(x, y):
    assert x[0] == y[0]
    assert x[1] == y[1]
    assert (x[2] - y[2]) < 1e-15 * openmm.unit.kilojoule_per_mole


def _compare_torsion_forces(force1, force2):
    sorted1 = _create_torsion_dict(torsion_force=force1)
    sorted2 = _create_torsion_dict(torsion_force=force2)

    assert sum(len(v) for v in sorted1.values()) == force1.getNumTorsions()
    assert sum(len(v) for v in sorted2.values()) == force2.getNumTorsions()
    assert len(sorted1) == len(sorted2)

    for key in sorted1:
        for i in range(len(sorted1[key])):
            _compare_individual_torsions(sorted1[key][i], sorted2[key][i])


@requires_package("openmm")
def _compare_bond_forces(force1, force2):
    assert force1.getNumBonds() == force2.getNumBonds()

    bonds1 = _create_bond_dict(force1)
    bonds2 = _create_bond_dict(force2)

    for key in bonds1:
        length_diff = bonds2[key][0] - bonds1[key][0]
        assert abs(length_diff) < 1e-15 * openmm.unit.nanometer, f"Bond lengths differ by {length_diff}"
        k_diff = bonds2[key][1] - bonds1[key][1]
        assert abs(k_diff) < 1e-9 * kj_nm2_mol, f"bond k differ by {k_diff}"


@requires_package("openmm")
def _compare_angle_forces(force1, force2):
    assert force1.getNumAngles() == force2.getNumAngles()

    angles1 = _create_angle_dict(force1)
    angles2 = _create_angle_dict(force2)

    for key in angles1:
        angle_diff = angles2[key][0] - angles1[key][0]
        assert abs(angle_diff) < 1e-15 * openmm.unit.radian, f"angles differ by {angle_diff}"
        k_diff = angles2[key][1] - angles1[key][1]
        assert abs(k_diff) < 1e-10 * kj_rad2_mol, f"angle k differ by {k_diff}"


def _compare_nonbonded_settings(force1, force2):
    for attr in dir(force1):
        if not attr.startswith("get") or attr in [
            "getExceptionParameterOffset",
            "getExceptionParameters",
            "getGlobalParameterDefaultValue",
            "getGlobalParameterName",
            "getLJPMEParametersInContext",
            "getPMEParametersInContext",
            "getParticleParameterOffset",
            "getParticleParameters",
            "getForceGroup",
        ]:
            continue
        assert getattr(force1, attr)() == getattr(force2, attr)(), attr


@requires_package("openmm")
def _compare_nonbonded_parameters(force1, force2):
    assert force1.getNumParticles() == force2.getNumParticles(), "found different number of particles"

    for i in range(force1.getNumParticles()):
        q1, sig1, eps1 = force1.getParticleParameters(i)
        q2, sig2, eps2 = force2.getParticleParameters(i)
        assert abs(q2 - q1) < 1e-8 * openmm.unit.elementary_charge, f"charge mismatch in particle {i}: {q1} vs {q2}"
        assert abs(sig2 - sig1) < 1e-12 * openmm.unit.nanometer, f"sigma mismatch in particle {i}: {sig1} vs {sig2}"
        assert abs(eps2 - eps1) < 1e-12 * openmm.unit.kilojoule_per_mole, (
            f"epsilon mismatch in particle {i}: {eps1} vs {eps2}"
        )


@requires_package("openmm")
def _compare_exceptions(force1, force2):
    assert force1.getNumExceptions() == force2.getNumExceptions(), "found different number of exceptions"

    for i in range(force1.getNumExceptions()):
        _, _, q1, sig1, eps1 = force1.getExceptionParameters(i)
        _, _, q2, sig2, eps2 = force2.getExceptionParameters(i)
        assert abs(q2 - q1) < 1e-12 * openmm.unit.elementary_charge**2, f"charge mismatch in exception {i}"
        assert abs(sig2 - sig1) < 1e-12 * openmm.unit.nanometer, f"sigma mismatch in exception {i}"
        assert abs(eps2 - eps1) < 1e-12 * openmm.unit.kilojoule_per_mole, f"epsilon mismatch in exception {i}"


@requires_package("openmm")
def _get_force(openmm_sys: "openmm.System", force_type):
    forces = [f for f in openmm_sys.getForces() if type(f) is force_type]

    if len(forces) > 1:
        raise NotImplementedError("Not yet able to process duplicate forces types")
    return forces[0]



import warnings

import pytest
from openff.toolkit import Quantity, Topology

from openff.interchange.components.mdconfig import (
    MDConfig,
    get_intermol_defaults,
    get_smirnoff_defaults,
)
from openff.interchange.constants import _PME
from openff.interchange.warnings import SwitchingFunctionNotImplementedWarning


@pytest.fixture
def system_no_constraints(sage_unconstrained, basic_top):
    return sage_unconstrained.create_interchange(basic_top)


@pytest.fixture
def rigid_water_box(sage, water):
    topology = water.to_topology()
    topology.box_vectors = Quantity([5, 5, 5], "nanometer")
    return sage.create_interchange(topology)


@pytest.fixture
def constrained_ligand_rigid_water_box(sage, basic_top, water):
    topology = Topology(basic_top)
    topology.add_molecule(water)
    return sage.create_interchange(topology)


@pytest.fixture
def unconstrained_ligand_rigid_water_box(sage_unconstrained, basic_top, water):
    topology = Topology(basic_top)
    topology.add_molecule(water)
    return sage_unconstrained.create_interchange(topology)


def parse_mdp(file: str) -> dict[str, str]:
    """Naively parse an MDP file into a dict structure."""
    options = dict()

    with open(file) as f:
        for line in f.readlines():
            line, _semicolon, _comment = line.partition(";")

            split = [token.replace("-", "") for token in line.split()]

            if len(split) == 0:
                continue

            elif len(split) == 3:
                assert split[1] == "="

                options[split[0].lower()] = split[2].lower()

            else:
                raise Exception

    return options


def parse_sander(file: str) -> dict[str, dict | str]:
    """Naively parse (sections of) a sander input file into a dict structure."""
    options: dict[str, dict | str] = dict()
    current_level = options

    with open(file) as f:
        for number, line in enumerate(f.readlines()):
            if number == 0:
                continue

            line = line.strip().replace(" ", "").replace(",", "").lower()

            if line.startswith("/"):
                continue

            if line.startswith("&"):
                current_section = line[1:]
                current_level[current_section] = dict()
                current_level = current_level[current_section]  # type: ignore[assignment]
                continue

            if "=" in line:
                key, value = line.split("=")
                current_level[key.lower()] = value
                continue

            raise ValueError(f"Unexpected content {line} on line {number} of {file}.")

    return options


class TestMDConfigFromInterchange:
    @pytest.mark.parametrize("periodic", [True, False])
    @pytest.mark.parametrize("switch", [True, False])
    def test_from_interchange(self, sage, basic_top, switch, periodic):
        from packaging.version import Version

        from openff.interchange import Interchange

        # After openff-toolkit 0.14.2, this should mean
        # periodic_method="cutoff", nonperiodic-method="no-cutoff"
        assert sage["vdW"].version >= Version("0.3.0")

        if not switch:
            sage["vdW"].switch_width = Quantity(0.0, "nanometer")

        if not periodic:
            basic_top.box_vectors = None

        interchange = Interchange.from_smirnoff(sage, basic_top)
        config = MDConfig.from_interchange(interchange)

        if switch:
            assert config.switching_function
            assert config.switching_distance.m_as("nanometer") == pytest.approx(0.8)
        else:
            assert not config.switching_function
            # No need to check the value of `switching_distance` ... right?

        if periodic:
            assert config.vdw_method == "cutoff"
        else:
            assert config.vdw_method == "no-cutoff"


class TestSMIRNOFFDefaults:
    @pytest.mark.parametrize("periodic", [True, False])
    def test_apply_smirnoff_defaults(self, sage, basic_top, periodic):
        from openff.interchange import Interchange

        interchange = Interchange.from_smirnoff(sage, basic_top)
        get_smirnoff_defaults(periodic=True).apply(interchange)

        for attr, value in zip(["cutoff", "switch_width"], [0.9, 0.1]):
            assert getattr(interchange["vdW"], attr).m_as(
                "nanometer",
            ) == pytest.approx(value)

        assert interchange["vdW"].periodic_method == "cutoff"
        assert interchange["vdW"].nonperiodic_method == "no-cutoff"

        if periodic:
            assert interchange["Electrostatics"].periodic_potential == _PME
        else:
            assert interchange["Electrostatics"].nonperiodic_potential == "Coulomb"


class TestIntermolDefaults:
    @pytest.mark.parametrize("periodic", [True, False])
    def test_write_mdp(self, periodic):
        """
        https://github.com/shirtsgroup/InterMol/blob/master/intermol/tests/gromacs/grompp_vacuum.mdp
        https://github.com/shirtsgroup/InterMol/blob/master/intermol/tests/gromacs/grompp.mdp
        """

        get_intermol_defaults(
            periodic=periodic,
        ).write_mdp_file("tmp.mdp")

        options = parse_mdp("tmp.mdp")

        assert options["pbc"] == "xyz" if periodic else "no"

        assert options["coulombtype"] == "pme" if periodic else "cutoff"
        assert options["rcoulomb"] == "0.9" if periodic else "2.0"
        assert options["coulombmodifier"] == "none"

        assert options["vdwtype"] == "cutoff"
        assert options["rvdw"] == "0.9"
        assert options["vdwmodifier"] == "none"

        # This is set to "no" in the vacuum file, though it probably doesn't matter
        assert options["dispcorr"] == "ener"

        assert options["constraints"] == "none"

    def test_apply_intermol_defaults(self, sage, basic_top):
        from openff.interchange import Interchange

        interchange = Interchange.from_smirnoff(sage, basic_top)
        get_intermol_defaults(periodic=True).apply(interchange)

        assert interchange["vdW"].switch_width.m == 0.0

        for key in ["vdW", "Electrostatics"]:
            assert interchange[key].cutoff.m_as("nanometer") == pytest.approx(0.9)


class TestWriteSanderInput:
    def test_system_no_constraints(self, system_no_constraints):
        MDConfig.from_interchange(system_no_constraints).write_sander_input_file("1.in")

        options = parse_sander("1.in")

        assert options["cntrl"]["ntf"] == "1"
        assert options["cntrl"]["ntc"] == "1"

    def test_rigid_water_box(self, rigid_water_box):
        MDConfig.from_interchange(rigid_water_box).write_sander_input_file("2.in")

        options = parse_sander("2.in")

        # Unclear what's "supposed" to be the value here
        assert options["cntrl"]["ntf"] in ("2", "4")
        assert options["cntrl"]["ntc"] == "1"

    def test_constrained_ligand_rigid_water_box(
        self,
        constrained_ligand_rigid_water_box,
    ):
        MDConfig.from_interchange(
            constrained_ligand_rigid_water_box,
        ).write_sander_input_file("3.in")

        options = parse_sander("3.in")

        # Amber does not support this case (constrain all h-bonds and water angles, but no other angles),
        # this option seems to be the best one
        assert options["cntrl"]["ntf"] == "2"
        assert options["cntrl"]["ntc"] == "1"

    def test_unconstrained_ligand_rigid_water_box(
        self,
        unconstrained_ligand_rigid_water_box,
    ):
        MDConfig.from_interchange(
            unconstrained_ligand_rigid_water_box,
        ).write_sander_input_file("3.in")

        options = parse_sander("3.in")

        # Amber does not support this case (constrain wtaer bonds and water angles, but no other bonds or angles),
        # this option seems to be the best one
        assert options["cntrl"]["ntf"] == "2"
        assert options["cntrl"]["ntc"] == "1"

    def test_switching_function_warning(
        self,
        system_no_constraints,
    ):
        """Reproduce issue 745."""
        with pytest.warns(
            SwitchingFunctionNotImplementedWarning,
        ):
            MDConfig.from_interchange(
                system_no_constraints,
            ).write_sander_input_file("yes.in")

        # This should always be -1 (turned off) since Amber does not implement
        # a switching function on the potential
        assert parse_sander("yes.in")["cntrl"]["fswitch"] == "-1.0"

        system_no_constraints["vdW"].switch_width = Quantity(0.0, "nanometer")

        # this pattern forces warnings to become errors
        with warnings.catch_warnings():
            warnings.simplefilter("error")

            MDConfig.from_interchange(
                system_no_constraints,
            ).write_sander_input_file("no.in")

        assert parse_sander("no.in")["cntrl"]["fswitch"] == "-1.0"


class TestWriteLAMMPSInput:
    def test_switching_function_warning(
        self,
        system_no_constraints,
        tmp_path,
    ):
        with pytest.warns(
            SwitchingFunctionNotImplementedWarning,
        ):
            MDConfig.from_interchange(
                system_no_constraints,
            ).write_lammps_input(
                interchange=system_no_constraints,
                input_file=tmp_path / "yes.in",
            )

    def test_no_error_if_no_constraints(
        self,
        system_no_constraints,
        tmp_path,
    ):
        MDConfig.from_interchange(system_no_constraints).write_lammps_input(
            interchange=system_no_constraints,
            input_file=tmp_path / ".inp",
        )

    def test_error_if_unsupported_constrained(
        self,
        rigid_water_box,
        constrained_ligand_rigid_water_box,
        unconstrained_ligand_rigid_water_box,
        tmp_path,
    ):
        for system in [
            rigid_water_box,
            constrained_ligand_rigid_water_box,
            unconstrained_ligand_rigid_water_box,
        ]:
            with pytest.raises(
                NotImplementedError,
                match="unsupported constraints case",
            ):
                MDConfig.from_interchange(system).write_lammps_input(
                    system,
                    tmp_path / ".inp",
                )



"""
Common helpers for exporting virtual sites.
"""

import abc
from collections import defaultdict
from typing import TYPE_CHECKING

import numpy
from openff.toolkit import Quantity

from openff.interchange.exceptions import (
    MissingPositionsError,
    MissingVirtualSitesError,
)
from openff.interchange.models import BaseVirtualSiteKey
from openff.interchange.pydantic import _BaseModel

if TYPE_CHECKING:
    import openmm

    from openff.interchange import Interchange


class _OpenMMVirtualSite(_BaseModel, abc.ABC):
    particles: list[int]

    @abc.abstractmethod
    def to_openmm(self) -> "openmm.VirtualSite":
        """Create a (subclass of) openmm.VirtualSite"""
        raise NotImplementedError()


class _TwoParticleAverageSite(_OpenMMVirtualSite):
    weights: list[float]

    def to_openmm(self):
        import openmm

        return openmm.TwoParticleAverageSite(
            particle1=self.particles[0],
            particle2=self.particles[1],
            weight1=self.weights[0],
            weight2=self.weights[1],
        )


class _ThreeParticleAverageSite(_OpenMMVirtualSite):
    weights: list[float]

    def to_openmm(self):
        import openmm

        # These arguments can't be named because (vague SWIG reasons)
        return openmm.ThreeParticleAverageSite(
            self.particles[0],  # particle1
            self.particles[1],  # particle2
            self.particles[2],  # particle3
            self.weights[0],  # weight1
            self.weights[1],  # weight2
            self.weights[2],  # weight3
        )


def _virtual_site_parent_molecule_mapping(
    interchange: "Interchange",
) -> dict[BaseVirtualSiteKey, int]:
    """
    Map `VirtualSiteKey`s the index of the molecule they belong to.

    Parameters
    ----------
    interchange
        The interchange object to get the mapping from.

    Returns
    -------
    mapping
        A dictionary mapping virtual site keys to the index of the molecule they belong to.

    """
    mapping: dict[BaseVirtualSiteKey, int] = dict()

    if "VirtualSites" not in interchange.collections:
        return mapping

    # TODO: This implicitly assumes the ordering of virtual sites is defined by
    # how they are presented in the iterator; this may cause problems when a
    # molecule (a large ligand? polymer?) has many virtual sites
    for virtual_site_key in interchange["VirtualSites"].key_map:
        assert isinstance(virtual_site_key, BaseVirtualSiteKey), type(virtual_site_key)

        parent_atom_index = virtual_site_key.orientation_atom_indices[0]

        parent_atom = interchange.topology.atom(parent_atom_index)

        parent_molecule = parent_atom.molecule

        mapping[virtual_site_key] = interchange.topology.molecule_index(parent_molecule)

    return mapping


def _get_molecule_virtual_site_map(interchange: "Interchange") -> defaultdict[int, list[BaseVirtualSiteKey]]:
    virtual_site_molecule_map = _virtual_site_parent_molecule_mapping(interchange)

    molecule_virtual_site_map = defaultdict(list)

    for virtual_site, molecule_index in virtual_site_molecule_map.items():
        molecule_virtual_site_map[molecule_index].append(virtual_site)

    return molecule_virtual_site_map


def get_positions_with_virtual_sites(
    interchange: "Interchange",
    collate: bool = False,
    use_zeros: bool = False,
) -> Quantity:
    """Return the positions of all particles (atoms and virtual sites)."""
    from openff.interchange.interop.common import _build_particle_map

    if interchange.positions is None:
        raise MissingPositionsError(
            f"Positions are required, found {interchange.positions=}.",
        )

    if "VirtualSites" not in interchange.collections:
        raise MissingVirtualSitesError()

    if len(interchange["VirtualSites"].key_map) == 0:
        raise MissingVirtualSitesError()

    # map of molecule index to *list* of virtual site keys contained therein
    molecule_virtual_site_map: defaultdict[int, list[BaseVirtualSiteKey]] = defaultdict(
        list,
    )

    virtual_site_molecule_map = _virtual_site_parent_molecule_mapping(interchange)

    for virtual_site, molecule_index in virtual_site_molecule_map.items():
        molecule_virtual_site_map[molecule_index].append(virtual_site)

    if "VirtualSites" in interchange.collections:
        if use_zeros:
            # TODO: Consider removing this behavior
            virtual_site_positions = numpy.zeros(
                (
                    len(interchange["VirtualSites"].key_map),
                    3,
                ),
            )
        else:
            from openff.interchange.smirnoff._virtual_sites import _generate_positions

            virtual_site_positions = _generate_positions(
                interchange,
                interchange["VirtualSites"],
            )

        if collate:
            # for i.e. a 4-site water dimer, these would be

            # {
            #   0: 0,
            #   1: 1,
            #   2: 2,
            #   3: 3,
            #   4: 4,
            #   5: 5,
            #   VirtualSiteKey with atom indices None: 6,
            #   VirtualSiteKey with atom indices None: 7,
            #   }
            uncollated = _build_particle_map(
                interchange=interchange,
                molecule_virtual_site_map=molecule_virtual_site_map,
                collate=False,
            )

            # {
            #   0: 0,
            #   1: 1,
            #   2: 2,
            #   VirtualSiteKey with atom indices None: 3,
            #   3: 4:
            #   4: 5,
            #   5: 6,
            #   VirtualSiteKey with atom indices None: 7,
            #   }
            collated = _build_particle_map(
                interchange=interchange,
                molecule_virtual_site_map=molecule_virtual_site_map,
                collate=True,
            )

            # and so the mapping from collated to uncollated would be
            # [0, 1, 2, 6, 3, 4, 5, 7]
            uncollated_to_collated_mapping = [uncollated[key] for key in collated]

            return numpy.concatenate(
                [
                    interchange.positions,
                    virtual_site_positions,
                ],
            )[uncollated_to_collated_mapping]

        else:
            # could just pass it through a [0, 1, 3, 4, ...] mapping
            # but that seems like a waste

            return numpy.concatenate(
                [
                    interchange.positions,
                    virtual_site_positions,
                ],
            )
    else:
        return interchange.positions


def _get_separation_by_atom_indices(
    interchange: "Interchange",
    atom_indices: tuple[int, ...],
    prioritize_geometry: bool = False,
) -> Quantity:
    """
    Given indices of (two?) atoms, return the distance between them.

    A constraint distance is first searched for, then an equilibrium bond length.

    This is slow, but often necessary for converting virtual site "distances" to weighted
    averages (unitless) of orientation atom positions.
    """
    if prioritize_geometry:
        p1 = interchange.positions[atom_indices[1]]  # type: ignore[index]
        p0 = interchange.positions[atom_indices[0]]  # type: ignore[index]

        return p1 - p0

    if "Constraints" in interchange.collections:
        collection = interchange["Constraints"]

        for key in collection.key_map:
            if (key.atom_indices == atom_indices) or (key.atom_indices[::-1] == atom_indices):
                return collection.potentials[collection.key_map[key]].parameters["distance"]

    if "Bonds" in interchange.collections:
        collection = interchange["Bonds"]  # type: ignore[assignment]

        for key in collection.key_map:
            if (key.atom_indices == atom_indices) or (key.atom_indices[::-1] == atom_indices):
                return collection.potentials[collection.key_map[key]].parameters["length"]

    # Two heavy atoms may be on opposite ends of an angle, in which case it's still
    # possible to determine their separation as defined by the geometry of the force field
    if "Angles" in interchange.collections:
        collection = interchange["Angles"]  # type: ignore[assignment]

        index0 = atom_indices[0]
        index1 = atom_indices[1]
        for key in collection.key_map:
            if (key.atom_indices[0] == index0 and key.atom_indices[2] == index1) or (
                key.atom_indices[2] == index0 and key.atom_indices[0] == index1
            ):
                gamma = collection.potentials[collection.key_map[key]].parameters["angle"]

                a = _get_separation_by_atom_indices(
                    interchange,
                    (key.atom_indices[0], key.atom_indices[1]),
                )

                b = _get_separation_by_atom_indices(
                    interchange,
                    (key.atom_indices[1], key.atom_indices[2]),
                )

                a = a.m_as("nanometer")
                b = b.m_as("nanometer")
                gamma = gamma.m_as("radian")

                # law of cosines
                c2 = a**2 + b**2 - 2 * a * b * numpy.cos(gamma)

                return Quantity(
                    c2**0.5,
                    "nanometer",
                )

    raise ValueError(f"Could not find distance between atoms {atom_indices}")



"""Utilities for processing and interfacing with the OpenFF Toolkit."""

from collections import defaultdict
from functools import lru_cache
from typing import TYPE_CHECKING

import networkx
import numpy
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.toolkit.topology._mm_molecule import _SimpleMolecule
from openff.toolkit.typing.engines.smirnoff.parameters import ParameterHandler, VirtualSiteHandler
from openff.toolkit.utils.collections import ValidatedList
from openff.utilities.utilities import has_package

from openff.interchange.models import ImportedVirtualSiteKey, PotentialKey

if has_package("openmm") or TYPE_CHECKING:
    import openmm.app


_IDIVF_1 = Quantity(1.0, "dimensionless")
_PERIODICITIES = {
    1: Quantity(1, "dimensionless"),
    2: Quantity(2, "dimensionless"),
    3: Quantity(3, "dimensionless"),
    4: Quantity(4, "dimensionless"),
    5: Quantity(5, "dimensionless"),
    6: Quantity(6, "dimensionless"),
}


def _get_num_h_bonds(topology: "Topology") -> int:
    """Get the number of (covalent) bonds containing a hydrogen atom."""
    n_bonds_containing_hydrogen = 0

    for bond in topology.bonds:
        if 1 in (bond.atom1.atomic_number, bond.atom2.atomic_number):
            n_bonds_containing_hydrogen += 1

    return n_bonds_containing_hydrogen


def _get_14_pairs(topology_or_molecule: Topology | Molecule):
    """Generate tuples of atom pairs, including symmetric duplicates."""
    # TODO: A replacement of Topology.nth_degree_neighbors in the toolkit
    #       may implement this in the future.
    for bond in topology_or_molecule.bonds:
        atom_i = bond.atom1
        atom_j = bond.atom2
        for atom_i_partner in atom_i.bonded_atoms:
            for atom_j_partner in atom_j.bonded_atoms:
                if atom_i_partner == atom_j_partner:
                    continue

                if atom_i_partner in atom_j_partner.bonded_atoms:
                    continue

                if atom_j_partner in atom_i_partner.bonded_atoms:
                    continue

                if {*atom_i_partner.bonded_atoms}.intersection(
                    {*atom_j_partner.bonded_atoms},
                ):
                    continue

                else:
                    atom_i_partner_index = topology_or_molecule.atom_index(
                        atom_i_partner,
                    )
                    atom_j_partner_index = topology_or_molecule.atom_index(
                        atom_j_partner,
                    )
                    if atom_i_partner_index > atom_j_partner_index:
                        yield (atom_j_partner, atom_i_partner)
                    else:
                        yield (atom_i_partner, atom_j_partner)


def _validated_list_to_array(validated_list: "ValidatedList") -> Quantity:
    unit_ = validated_list[0].units
    return Quantity(numpy.asarray([val.m for val in validated_list]), unit_)


def _combine_topologies(topology1: Topology, topology2: Topology) -> Topology:
    topology1_ = Topology(other=topology1)
    topology2_ = Topology(other=topology2)

    for molecule in topology2_.molecules:
        topology1_.add_molecule(molecule)

    return topology1_


def _check_electrostatics_handlers(force_field: "ForceField") -> bool:
    """
    Return whether or not this ForceField should have an Electrostatics tag.
    """
    # Manually-curated list of names of ParameterHandler classes that are expected
    # to assign/modify partial charges
    partial_charge_handlers = [
        "LibraryCharges",
        "ToolkitAM1BCC",
        "ChargeIncrementModel",
        "VirtualSites",
        "GBSA",
    ]

    # A more robust solution would probably take place late in the parameterization
    # process, but this solution should cover a vast majority of cases with minimal
    # complexity. Most notably this will *not* behave well with
    #   * parameter handler plugins
    #   * future additions to -built-in handlers
    #   * handlers that _could_ assign partial charges but happen to not assign
    #       any for some particular topology

    for parameter_handler_name in force_field.registered_parameter_handlers:
        if parameter_handler_name in partial_charge_handlers:
            return True

    return False


def _simple_topology_from_openmm(
    openmm_topology: "openmm.app.Topology",
    system: openmm.System | None = None,
) -> Topology:
    """
    Convert an OpenMM Topology into an OpenFF Topology consisting **only** of so-called `_SimpleMolecule`s.

    Arguments
    ---------
    openmm_topology
        The OpenMM Topology to convert.
    system
        The OpenMM System associated with the topology. Only needed if there are virtual sites in the topology.
    """
    # TODO: Splice in fully-defined OpenFF `Molecule`s?

    graph = networkx.Graph()

    virtual_sites: list[ImportedVirtualSiteKey] = list()

    # map indices of OpenMM system with virtual site to topology indices, or virtual site keys,
    # in associated OpenFF topology, since stripping out virtual sites offsets particle indices
    openmm_openff_particle_map: dict[int, int | ImportedVirtualSiteKey] = dict()

    for atom in openmm_topology.atoms():
        if atom.element is None:
            assert isinstance(
                system,
                openmm.System,
            ), "`system` argument required if virtual sites are present in the topology"
            try:
                # assume ThreeParticleAverageSite for now
                orientation_atom_indices = [
                    openmm_openff_particle_map[system.getVirtualSite(atom.index).getParticle(i)] for i in range(3)
                ]
            except openmm.OpenMMException as error:
                if "This particle is not a virtual site" in str(error):
                    raise ValueError(
                        "Particle ordering mismatch between OpenMM system and topology. "
                        f"Look at particle {atom.index} in the topology.",
                    ) from error
            virtual_sites.append(
                ImportedVirtualSiteKey(
                    orientation_atom_indices=orientation_atom_indices,
                    name=atom.name,
                    type="ThreeParticleAverageSite",
                ),
            )

            # TODO: This will break if virtual sites are not after their parent/orientation atoms
            openmm_openff_particle_map[atom.index] = virtual_sites[-1]

        else:
            graph.add_node(
                atom.index,
                atomic_number=getattr(atom.element, "atomic_number", 0),
                name=atom.name,
                residue_name=atom.residue.name,
                # Note that residue number is mapped to residue.id here. The use of id vs. number
                # varies in other packages and the convention for the OpenFF-OpenMM interconversion
                # is recorded at
                # https://docs.openforcefield.org/projects/toolkit/en/0.15.1/users/molecule_conversion.html
                residue_number=atom.residue.id,
                insertion_code=atom.residue.insertionCode,
                chain_id=atom.residue.chain.id,
            )

            openmm_openff_particle_map[atom.index] = atom.index - len(virtual_sites)

    for bond in openmm_topology.bonds():
        graph.add_edge(
            bond.atom1.index,
            bond.atom2.index,
        )

    topology = _simple_topology_from_graph(graph)

    topology._molecule_virtual_site_map = defaultdict(list)

    # TODO: This iteration strategy scales horribly with system size - need to refactor -
    #       since looking up topology atom indices is slow. It's probably repetitive to
    #       look up the molecule index over and over again
    for particle in virtual_sites:
        molecule_index = topology.molecule_index(topology.atom(particle.orientation_atom_indices[0]).molecule)

        topology._molecule_virtual_site_map[molecule_index].append(particle)

    topology._particle_map = openmm_openff_particle_map

    return topology


def _simple_topology_from_graph(graph: networkx.Graph) -> Topology:
    """Convert a networkx Graph into an OpenFF Topology consisting only of `_SimpleMolecule`s."""
    topology = Topology()

    for component in networkx.connected_components(graph):
        subgraph = _reorder_subgraph(graph.subgraph(component))

        # Attempt to safeguard against the possibility that
        # the subgraphs are returned out of "atom order", like
        # if atoms in an later molecule have lesser atom indices
        # than this molecule
        #
        # Oct 2024 - need to add a test case for above?
        # these values are not necessarily equal because of virtual sites
        assert topology.n_atoms <= next(iter(subgraph.nodes))

        topology.add_molecule(_SimpleMolecule._from_subgraph(subgraph))

    return topology


def _reorder_subgraph(graph: networkx.Graph) -> networkx.Graph:
    """Ensure that the graph is ordered with ascending atoms."""
    new_graph = networkx.Graph()
    new_graph.add_nodes_from(sorted(graph.nodes(data=True)))
    new_graph.add_edges_from(graph.edges(data=True))

    return new_graph


# This is to re-implement:
#   https://github.com/openforcefield/openff-toolkit/blob/60014820e6a333bed04e8bf5181d177da066da4d/
#   openff/toolkit/typing/engines/smirnoff/parameters.py#L2509-L2515
# It doesn't seem ideal to assume that matching SMILES === isomorphism?
class _HashedMolecule(Molecule):
    def __hash__(self):
        return hash(self.to_smiles(mapped=True, explicit_hydrogens=True, isomeric=True))


def _assert_all_isomorphic(molecule_list: list[Molecule]) -> bool:
    hashed_molecules = {_HashedMolecule(molecule) for molecule in molecule_list}

    return len(hashed_molecules) == len(molecule_list)


def _lookup_virtual_site_parameter(
    parameter_handler: VirtualSiteHandler,
    smirks: str,
    name: str,
    match: str,
) -> VirtualSiteHandler.VirtualSiteType:
    """
    Given some attributes, look up a virtual site parameter.

    The toolkit does not reliably look up `VirtualSiteType`s when SMIRKS are not unique,
    which is valid for some virtual site use cases.
    https://github.com/openforcefield/openff-toolkit/issues/1847

    """
    if not isinstance(parameter_handler, VirtualSiteHandler):
        raise NotImplementedError("Only VirtualSiteHandler is currently supported.")

    for virtual_site_type in parameter_handler.parameters:
        if virtual_site_type.smirks == smirks and virtual_site_type.name == name and virtual_site_type.match == match:
            return virtual_site_type
    else:
        raise ValueError(
            f"No VirtualSiteType found with {smirks=}, name={name=}, and match={match=}.",
        )


@lru_cache
def _cache_angle_parameter_lookup(
    potential_key: PotentialKey,
    parameter_handler: ParameterHandler,
) -> dict[str, Quantity]:
    parameter = parameter_handler.parameters[potential_key.id]

    return {parameter_name: getattr(parameter, parameter_name) for parameter_name in ["k", "angle"]}


@lru_cache
def _cache_torsion_parameter_lookup(
    potential_key: PotentialKey,
    parameter_handler: ParameterHandler,
    idivf: float | None = None,
) -> dict[str, Quantity]:
    smirks = potential_key.id
    n = potential_key.mult
    parameter = parameter_handler.parameters[smirks]

    if idivf is not None:
        # case of non-standard default_idivf in impropers
        _idivf = idivf
    elif parameter.idivf is None:
        # This appears to only come from imports
        _idivf = _IDIVF_1
    elif parameter.idivf[n] == 1.0:
        _idivf = _IDIVF_1
    else:
        _idivf = Quantity(parameter.idivf[n], "dimensionless")

    return {
        "k": parameter.k[n],
        "periodicity": _PERIODICITIES[parameter.periodicity[n]],
        "phase": parameter.phase[n],
        "idivf": _idivf,
    }



import pytest
from openff.toolkit import Molecule, Topology
from openff.toolkit.topology._mm_molecule import _SimpleMolecule
from openff.utilities.testing import skip_if_missing

from openff.interchange._tests import get_test_file_path
from openff.interchange.components.toolkit import (
    _check_electrostatics_handlers,
    _combine_topologies,
    _get_14_pairs,
    _get_num_h_bonds,
    _lookup_virtual_site_parameter,
    _simple_topology_from_openmm,
)


@pytest.fixture
def simple_methane():
    return _SimpleMolecule.from_molecule(Molecule.from_smiles("C"))


@pytest.fixture
def simple_water(water):
    return _SimpleMolecule.from_molecule(water)


def test_simple_topology_uniqueness(simple_methane, simple_water):
    topology = Topology.from_molecules(
        [
            simple_methane,
            simple_water,
            simple_methane,
            simple_methane,
            simple_water,
        ],
    )
    assert len(topology.identical_molecule_groups) == 2


class TestToolkitUtils:
    @pytest.mark.parametrize(
        ("smiles", "num_pairs"),
        [
            ("C#C", 1),
            ("CCO", 12),
            ("C1=CC=CC=C1", 24),
            ("C=1=C=C1", 0),
            ("C=1=C=C=C1", 0),
            ("C=1(Cl)-C(Cl)=C1", 1),
            ("C=1=C(Cl)C(=C=1)Cl", 5),
        ],
    )
    def test_get_14_pairs(self, smiles, num_pairs):
        mol = Molecule.from_smiles(smiles)
        assert len([*_get_14_pairs(mol)]) == num_pairs
        assert len([*_get_14_pairs(mol.to_topology())]) == num_pairs

    def test_check_electrostatics_handlers(self, tip3p):
        tip3p.deregister_parameter_handler("Electrostatics")

        assert _check_electrostatics_handlers(tip3p)

        tip3p.deregister_parameter_handler("LibraryCharges")

        assert not _check_electrostatics_handlers(tip3p)

    @pytest.mark.parametrize(
        ("smiles", "num_h_bonds"),
        [("C", 4), ("C#C", 2), ("O", 2)],
    )
    def test_get_num_h_bonds(self, smiles, num_h_bonds):
        topology = Molecule.from_smiles(smiles).to_topology()
        assert _get_num_h_bonds(topology) == num_h_bonds, smiles

    def test_combine_topologies(self, water):
        ethanol = Molecule.from_smiles("CCO")
        ethanol.name = "ETH"
        ethanol_topology = ethanol.to_topology()

        water.name = "WAT"
        water_topology = water.to_topology()

        combined = _combine_topologies(ethanol_topology, water_topology)

        for attr in (
            "atoms",
            "bonds",
        ):
            attr = "n_" + attr
            assert getattr(combined, attr) == getattr(ethanol_topology, attr) + getattr(
                water_topology,
                attr,
            )

    @skip_if_missing("openmm")
    def test_simple_topology_from_openmm(self):
        simple_topology = _simple_topology_from_openmm(
            Topology.from_molecules(
                [
                    Molecule.from_smiles("O"),
                    Molecule.from_smiles("CCO"),
                ],
            ).to_openmm(),
        )

        assert all(isinstance(molecule, _SimpleMolecule) for molecule in simple_topology.molecules)

        assert sorted(molecule.n_atoms for molecule in simple_topology.molecules) == [
            3,
            9,
        ]

    def test_simple_subgraph_atom_ordering(self):
        """
        Test that simple molecules created from subgraphs use nodes in ascending
        order (like you'd always do for atom indices)
        """
        pytest.importorskip("openmm")

        import openmm.app

        crazy_water = openmm.app.PDBFile(
            get_test_file_path("copied_reordered_water.pdb").as_posix(),
        )

        expected_atomic_numbers = [
            8,
            1,
            1,
            1,
            8,
            1,
            1,
            1,
            8,
            1,
            8,
            1,
        ]

        simple_topology = _simple_topology_from_openmm(crazy_water.topology)

        for atom, expected_atomic_number in zip(
            simple_topology.atoms,
            expected_atomic_numbers,
        ):
            assert atom.atomic_number == expected_atomic_number


def test_lookup_virtual_site_parameter(
    sage,
    tip4p,
    sage_with_two_virtual_sites_same_smirks,
):
    with pytest.raises(
        NotImplementedError,
    ):
        _lookup_virtual_site_parameter(
            sage["LibraryCharges"],
            smirks="",
            name="",
            match="",
        )

    assert _lookup_virtual_site_parameter(
        parameter_handler=tip4p["VirtualSites"],
        smirks="[#1:2]-[#8X2H2+0:1]-[#1:3]",
        name="EP",
        match="once",
    )

    ep1 = _lookup_virtual_site_parameter(
        sage_with_two_virtual_sites_same_smirks["VirtualSites"],
        smirks="[#7:1](-[*:2])(-[*:3])-[*:4]",
        name="EP1",
        match="once",
    )
    assert ep1.distance.m_as("nanometer") == -0.5

    ep2 = _lookup_virtual_site_parameter(
        sage_with_two_virtual_sites_same_smirks["VirtualSites"],
        smirks="[#7:1](-[*:2])(-[*:3])-[*:4]",
        name="EP2",
        match="once",
    )
    assert ep2.distance.m_as("nanometer") == 1.5

    with pytest.raises(
        ValueError,
        match=r"No VirtualSiteType found with .*EP3",
    ):
        _lookup_virtual_site_parameter(
            sage_with_two_virtual_sites_same_smirks["VirtualSites"],
            smirks="[#7:1](-[*:2])(-[*:3])-[*:4]",
            name="EP3",
            match="once",
        )



import os.path
import tempfile
import time

import openmm.app
import openmm.unit
import polars
from openff.toolkit import ForceField, Molecule, Topology
from tqdm import tqdm

from openff.interchange import __version__

data_frame_file = f"{__version__}/protein_in_water.csv"

if os.path.exists(data_frame_file):
    exit("CSV file already exists, exiting")

protein_force_field = ForceField("openff_no_water-3.0.0-alpha0.offxml", "tip3p.offxml")

pdb_topology = Topology.from_pdb("ctr148a-2KO1-model-1.pdb")

modeller = openmm.app.Modeller(
    topology=pdb_topology.to_openmm(),
    positions=pdb_topology.get_positions().to_openmm(),
)


modeller.addSolvent(
    forcefield=openmm.app.ForceField("amber99sb.xml", "tip3p.xml"),
    model="tip3p",
    padding=1 * openmm.unit.nanometer,
)

topology = Topology.from_openmm(
    modeller.topology,
    unique_molecules=[*list({*pdb_topology.molecules}), Molecule.from_smiles("O"), Molecule.from_smiles("[Cl-]")],
    positions=modeller.getPositions(),
)

interchange = protein_force_field.create_interchange(topology)

t0 = time.perf_counter()

interchange.to_openmm()

print(f"Time to create OpenMM System: {time.perf_counter() - t0} seconds")


times = list()

with tempfile.TemporaryDirectory() as tmpdir:
    for index in tqdm(range(10), desc="Ligand in water benchmark"):
        openmm_start = time.perf_counter()
        interchange.to_openmm()
        openmm_end = time.perf_counter()

        gromacs_start = time.perf_counter()
        interchange.to_gromacs(prefix=f"{tmpdir}/temp")
        gromacs_end = time.perf_counter()

        times.append(
            {
                "iteration": index,
                "openmm_time": openmm_end - openmm_start,
                "gromacs_time": gromacs_end - gromacs_start,
            },
        )

data_frame = polars.from_dicts(
    data=times,
    schema={
        "iteration": polars.Int64,
        "openmm_time": polars.Float64,
        "gromacs_time": polars.Float64,
    },
)

data_frame.write_csv(data_frame_file)



"""
Helper functions for exporting virutal sites to GROMACS.
"""

import numpy
from openff.toolkit import Quantity

from openff.interchange import Interchange
from openff.interchange.interop._virtual_sites import _get_separation_by_atom_indices
from openff.interchange.interop.gromacs.models.models import (
    GROMACSVirtualSite,
    GROMACSVirtualSite2,
    GROMACSVirtualSite3,
    GROMACSVirtualSite3fad,
)
from openff.interchange.models import BaseVirtualSiteKey
from openff.interchange.smirnoff._virtual_sites import (
    _BondChargeVirtualSite,
    _DivalentLonePairVirtualSite,
    _MonovalentLonePairVirtualSite,
    _VirtualSite,
)


def _create_gromacs_virtual_site(
    interchange: Interchange,
    virtual_site: "_VirtualSite",
    virtual_site_key: BaseVirtualSiteKey,
    particle_map: dict[int | BaseVirtualSiteKey, int],
) -> GROMACSVirtualSite:
    # Orientation atom indices are topology indices, but here they need to be indexed as molecule
    # indices. Store the difference between an orientation atom's molecule and topology indices.
    # (It can probably be any of the orientation atoms.)
    parent_atom = interchange.topology.atom(
        virtual_site_key.orientation_atom_indices[0],
    )

    # This lookup scales poorly with system size, but it's not clear how to work around the
    # tool's ~O(N) scaling of topology lookups
    offset = interchange.topology.atom_index(
        parent_atom,
    ) - parent_atom.molecule.atom_index(parent_atom)

    # These are GROMACS "molecule" indices, already mapped back from the topology on to the molecule
    gromacs_indices: list[int] = [
        particle_map[openff_index] - offset + 1 for openff_index in virtual_site.orientations
    ]

    if isinstance(virtual_site, _BondChargeVirtualSite):
        separation = _get_separation_by_atom_indices(
            interchange=interchange,
            atom_indices=virtual_site.orientations,
        )
        distance = virtual_site.distance

        ratio = (distance / separation).m_as("dimensionless")

        # TODO: This will create name collisions in many molecules
        return GROMACSVirtualSite2(
            name=virtual_site_key.name,
            site=particle_map[virtual_site_key] - offset + 1,
            orientation_atoms=gromacs_indices,
            a=-1.0 * ratio,  # this is basically w2 in OpenMM jargon
        )

    if isinstance(virtual_site, _DivalentLonePairVirtualSite):
        r12 = _get_separation_by_atom_indices(
            interchange=interchange,
            atom_indices=virtual_site.orientations[:2],
        )
        r13 = _get_separation_by_atom_indices(
            interchange=interchange,
            atom_indices=(virtual_site.orientations[0], virtual_site.orientations[2]),
        )

        distance = virtual_site.distance

        # TODO: Test r12 != r13, prima facia the math also applies, probably need
        #       a more direct way to get r1mid
        if r12 == r13 and float(virtual_site.out_of_plane_angle.m) == 0.0:
            r23 = _get_separation_by_atom_indices(
                interchange=interchange,
                atom_indices=virtual_site.orientations[1:],
            )

            theta = Quantity(
                numpy.arccos(
                    (r23**2 - r12**2 - r13**2) / (-2 * r12 * r13),
                ),
                "radian",
            )

            r1mid = Quantity(numpy.cos(theta.m_as("radian") / 2) * r12.m_as("nanometer"), "nanometer")

            w1 = 1 + distance / r1mid

            return GROMACSVirtualSite3(
                name=virtual_site_key.name,
                site=particle_map[virtual_site_key] - offset + 1,
                orientation_atoms=gromacs_indices,
                a=(1 - w1) / 2,
                b=(1 - w1) / 2,
            )

        else:
            raise NotImplementedError()

    if isinstance(virtual_site, _MonovalentLonePairVirtualSite):
        if virtual_site.out_of_plane_angle != 0.0:
            raise NotImplementedError(
                "Non-zero out-of-plane angles not yet supported in GROMACS export.",
            )

        # In the plane of three atoms, GROMACS calls this 3fad and gives the example
        #
        # [ virtual_sites3 ]
        # ; Site  from               funct   theta      d
        # 5       1     2     3      3       120        0.5
        # https://manual.gromacs.org/current/reference-manual/topologies/particle-type.html

        return GROMACSVirtualSite3fad(
            name=virtual_site_key.name,
            site=particle_map[virtual_site_key] - offset + 1,
            orientation_atoms=gromacs_indices,
            theta=virtual_site.in_plane_angle.m_as("degree"),
            d=virtual_site.distance.m_as("nanometer"),
        )

    raise NotImplementedError()



"""Minimize energy using OpenMM."""

from typing import TYPE_CHECKING

from openff.toolkit import Quantity
from openff.utilities.utilities import requires_package

from openff.interchange.exceptions import MinimizationError, MissingPositionsError

if TYPE_CHECKING:
    from openff.interchange import Interchange


@requires_package("openmm")
def minimize_openmm(
    interchange: "Interchange",
    tolerance: Quantity,
    max_iterations: int,
) -> Quantity:
    """Minimize the energy of a system using OpenMM."""
    import openmm
    import openmm.unit
    from openff.units.openmm import from_openmm

    simulation = interchange.to_openmm_simulation(
        integrator=openmm.LangevinMiddleIntegrator(
            293.15 * openmm.unit.kelvin,
            1.0 / openmm.unit.picosecond,
            2.0 * openmm.unit.femtosecond,
        ),
        combine_nonbonded_forces=False,
    )

    simulation.context.computeVirtualSites()

    try:
        simulation.minimizeEnergy(
            tolerance=tolerance.to_openmm(),
            maxIterations=max_iterations,
        )

    except openmm.OpenMMException as error:
        if "Particle positions have not been set" in str(error):
            raise MissingPositionsError(
                f"Cannot minimize without positions. Found {interchange.positions=}.",
            ) from error
        else:
            raise MinimizationError("OpenMM Minimization failed.") from error

    # Assume that all virtual sites are placed at the _end_, so the 0th through
    # (number of atoms)th positions are the massive particles
    return from_openmm(
        simulation.context.getState(
            getPositions=True,
        ).getPositions(
            asNumpy=True,
        )[
            : interchange.positions.shape[0],  # type: ignore[union-attr]
            :,
        ],
    )



import numpy
import pytest
from openff.toolkit import Quantity
from pydantic import ValidationError

from openff.interchange._annotations import _BoxQuantity, _kJMolQuantity, _MassQuantity
from openff.interchange.models import _BaseModel


class TestAnnotations:
    def test_dimensionality_check_fails(self):
        class Model(_BaseModel):
            mass: _MassQuantity

        with pytest.raises(
            ValidationError,
            match=r"Dimensionality.*is not compatible with",
        ):
            Model(mass=Quantity(400.0, "kelvin"))

    def test_dimensionality_check_soft_passes(self):
        class Model(_BaseModel):
            mass: _MassQuantity

        assert Model(mass=Quantity(1e-3 / 6.02214076e23, "kg")).mass.m_as("amu") == pytest.approx(1.0)

    def test_unit_check_coerced(self):
        class Model(_BaseModel):
            dde: _kJMolQuantity

        assert Model(dde=Quantity(1.5, "kcal/mol")).dde.units == "kilojoule / mole"

    def test_dimensionality_check_passes(self):
        class Model(_BaseModel):
            dde: _kJMolQuantity

        assert Model(dde=Quantity(1.5, "kJ/mol")).dde.m_as("kJ/mol") == pytest.approx(1.5)


class TestBoxQuantity:
    def test_list_cast_to_nanometer_quantity_array(self):
        class M(_BaseModel):
            box: _BoxQuantity

        box = M(box=[2, 3, 4]).box

        assert isinstance(box, Quantity)
        assert str(box.units) == "nanometer"
        assert box.shape == (3, 3)

        numpy.testing.assert_allclose(box, box * numpy.eye(3))

    def test_ndarray_to_nanometer_quantity_array(self):
        class M(_BaseModel):
            box: _BoxQuantity

        box = M(box=numpy.array([[2, 0, 0], [0.1, 3, 0], [0.2, 0.5, 4]])).box

        assert isinstance(box, Quantity)
        assert str(box.units) == "nanometer"
        assert box.shape == (3, 3)

        assert box[0][0].m == pytest.approx(2)
        assert box[2][1].m == pytest.approx(0.5)
